/* $Revision: 1.68 $
 * Copyright 1994-2002 The MathWorks, Inc.
 *
 * File    : grt_main.c
 *
 * Abstract:
 *      A Generic "Real-Time (single tasking or pseudo-multitasking,
 *      statically allocated data)" main that runs under most
 *      operating systems.
 *
 *      This file may be a useful starting point when targeting a new
 *      processor or microcontroller.
 *
 *
 * Compiler specified defines:
 *	RT              - Required.
 *      MODEL=modelname - Required.
 *	NUMST=#         - Required. Number of sample times.
 *	NCSTATES=#      - Required. Number of continuous states.
 *      TID01EQ=1 or 0  - Optional. Only define to 1 if sample time task
 *                        id's 0 and 1 have equal rates.
 *      MULTITASKING    - Optional. (use MT for a synonym).
 *	SAVEFILE        - Optional (non-quoted) name of .mat file to create.
 *			  Default is <MODEL>.mat
 *      BORLAND         - Required if using Borland C/C++
 */

#include "tmwtypes.h"
#include "rtmodel.h"
#include "rt_sim.h"
#include "rt_logging.h"
#include "rt_nonfinite.h"

/* ST-RTL headers */ 
#include "common.h"
#include "krnl_main.h"
#include "ext_svr_rtl.h"

/* rtl_printf() */
#include <rtl_printf.h>
#define printf rtl_printf

/* Signal Handler header */
#ifdef BORLAND
#include <signal.h>
#include <float.h>
#endif

#include "ext_work.h"

/* External Mode header */
#ifdef EXT_MODE
#include "updown.h"
#endif

/*=========*
 * Defines *
 *=========*/

#ifndef TRUE
#define FALSE (0)
#define TRUE  (1)
#endif

#ifndef EXIT_FAILURE
#define EXIT_FAILURE  1
#endif
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS  0
#endif

#define QUOTE1(name) #name
#define QUOTE(name) QUOTE1(name)    /* need to expand name    */

#ifndef RT
# error "must define RT"
#endif

#ifndef MODEL
# error "must define MODEL"
#endif

#ifndef NUMST
# error "must define number of sample times, NUMST"
#endif

#ifndef NCSTATES
# error "must define NCSTATES"
#endif

#define RUN_FOREVER -1.0

static int dataLogInit = FALSE; /* Check if rt_StopDataLogging should execute. */
#ifdef EXT_MODE
static int_T counter[NUMST];
#endif

/**********************
 * External Variables *
 **********************/
extern volatile int_T disFac;

/*====================*
 * External functions *
 *====================*/
extern RT_MODEL *MODEL(void);

extern void MdlInitializeSizes(void);
extern void MdlInitializeSampleTimes(void);
extern void MdlStart(void);
extern void MdlOutputs(int_T tid);
extern void MdlUpdate(int_T tid);
extern void MdlTerminate(void);

#if NCSTATES > 0
  extern void rt_ODECreateIntegrationData(RTWSolverInfo *si);
  extern void rt_ODEUpdateContinuousStates(RTWSolverInfo *si);

# define rt_CreateIntegrationData(S) \
    rt_ODECreateIntegrationData(rtmGetRTWSolverInfo(S));
# define rt_UpdateContinuousStates(S) \
    rt_ODEUpdateContinuousStates(rtmGetRTWSolverInfo(S));
# else
# define rt_CreateIntegrationData(S)  \
      rtsiSetSolverName(rtmGetRTWSolverInfo(S),"FixedStepDiscrete");
# define rt_UpdateContinuousStates(S) \
      rtmSetT(S, rtsiGetSolverStopTime(rtmGetRTWSolverInfo(S)));
#endif


/*==================================*
 * Global data local to this module *
 *==================================*/

static struct {
  int_T    stopExecutionFlag;
  int_T    overrunFlags[NUMST];
  const char_T *errmsg;
} GBLbuf;


#ifdef EXT_MODE
#  define rtExtModeSingleTaskUpload(S)    \
   {                                      \
        int stIdx;                        \
        rtExtModeUploadCheckTrigger();    \
        for (stIdx=0; stIdx<NUMST; stIdx++) {                   \
            if (rtmIsSampleHit(S, stIdx, 0 /*unused*/)) {       \
                  /* Buffer signal point according to the Discarding Factor (disFac) */ \
                  if (++counter[stIdx] >= disFac) {   \
                        counter[stIdx] = 0;           \
                        /* If disFac = 0 all the packets are to be discarded. */    \
                        if (disFac)               \
		                 rtExtModeUpload(stIdx,rtmGetTaskTime(S,stIdx)); \
                  }                                             \
            }                                                   \
        }                                                       \
   }
#else
#  define rtExtModeSingleTaskUpload(S) /* Do nothing */
#endif

/*=================*
 * Local functions *
 *=================*/

#ifdef BORLAND
/* Implemented for BC++ only*/

typedef void (*fptr)(int, int);

/* Function: divideByZero =====================================================
 *
 * Abstract: Traps the error Division by zero and prints a warning
 *           Also catches other FP errors, but does not identify them
 *           specifically.
 */
void divideByZero(int sigName, int sigType)
{
    signal(SIGFPE, (fptr)divideByZero);
    if ((sigType == FPE_ZERODIVIDE)||(sigType == FPE_INTDIV0)){
        PRINT_W("*** Warning: Division by zero\n\n");
        return;
    }
    else{
        PRINT_W("*** Warning: Floating Point error\n\n");
        return;
    }
} /* end divideByZero */

#endif /* BORLAND */

#if !defined(MULTITASKING)  /* SINGLETASKING */

/* Function: rtOneStep ========================================================
 *
 * Abstract:
 *      Perform one step of the model. This function is modeled such that
 *      it could be called from an interrupt service routine (ISR) with minor
 *      modifications.
 */
static void rt_OneStep(RT_MODEL *S, int_T tid)
{
    real_T tnext;

    /***********************************************
     * Check and see if error status has been set  *
     ***********************************************/

    if (rtmGetErrorStatus(S) != NULL) {
        GBLbuf.stopExecutionFlag = 1;
        return;
    }

#ifdef EXT_MODE
 	rt_UploadServerWork_rtl(); 
#endif
    
    tnext = rt_SimGetNextSampleHit();
    rtsiSetSolverStopTime(rtmGetRTWSolverInfo(S),tnext);

    MdlOutputs(0);

    rtExtModeSingleTaskUpload(S);

    GBLbuf.errmsg = rt_UpdateTXYLogVars(rtmGetRTWLogInfo(S),
                                        rtmGetTPtr(S));
    if (GBLbuf.errmsg != NULL) {
        GBLbuf.stopExecutionFlag = 1;
        return;
    }

    MdlUpdate(0);
    rt_SimUpdateDiscreteTaskSampleHits(rtmGetNumSampleTimes(S),
                                       rtmGetTimingData(S),
                                       rtmGetSampleHitPtr(S),
                                       rtmGetTPtr(S));

    if (rtmGetSampleTime(S,0) == CONTINUOUS_SAMPLE_TIME) {
        rt_UpdateContinuousStates(S);
    }

    rtExtModeCheckEndTrigger();

} /* end rtOneStep */

#else /* MULTITASKING */

# if TID01EQ == 1
#  define FIRST_TID 1
# else
#  define FIRST_TID 0
# endif

/* Function: rtOneStep ========================================================
 *
 * Abstract:
 *      Perform one step of the model. This function is modeled such that
 *      it could be called from an interrupt service routine (ISR) with minor
 *      modifications.
 *
 *      This routine is modeled for use in a multitasking environment and
 *	therefore needs to be fully re-entrant when it is called from an
 *	interrupt service routine.
 *
 * Note:
 *      Error checking is provided which will only be used if this routine
 *      is attached to an interrupt.
 *
 */
static void rt_OneStep(RT_MODEL *S, int_T tid)
{
    real_T tnext;

    /*******************************************
     * Step the model for the base sample time *
     *******************************************/
     if (tid == 0) {
     
#ifdef EXT_MODE
 	  rt_UploadServerWork_rtl(); 
#endif
       tnext = rt_SimUpdateDiscreteEvents(rtmGetNumSampleTimes(S),
                                       rtmGetTimingData(S),
                                       rtmGetSampleHitPtr(S),
                                       rtmGetPerTaskSampleHitsPtr(S));
       rtsiSetSolverStopTime(rtmGetRTWSolverInfo(S),tnext);     

        MdlOutputs(FIRST_TID);

        rtExtModeUploadCheckTrigger();
#ifdef EXT_MODE     
        /* Buffer signal point according to the Discarding Factor (disFac) */
        if (++counter[tid] >= disFac) {
            counter[tid] = 0;
            /* If disFac = 0 all the packets are to be discarded. */
            if (disFac)
                  rtExtModeUpload(FIRST_TID,rtmGetTaskTime(S, FIRST_TID));		
        }     
#endif
        GBLbuf.errmsg = rt_UpdateTXYLogVars(rtmGetRTWLogInfo(S),
                                        rtmGetTPtr(S));
        if (GBLbuf.errmsg != NULL) {
            GBLbuf.stopExecutionFlag = 1;
            return;
        }

        MdlUpdate(FIRST_TID);

        if (rtmGetSampleTime(S,0) == CONTINUOUS_SAMPLE_TIME) {
             rt_UpdateContinuousStates(S);
        }
        else {
             rt_SimUpdateDiscreteTaskTime(rtmGetTPtr(S), 
                                     rtmGetTimingData(S), 0);
        }

#if FIRST_TID == 1
        rt_SimUpdateDiscreteTaskTime(rtmGetTPtr(S), rtmGetTimingData(S),1);
#endif
        rtExtModeCheckEndTrigger();
    } else {

    /*********************************************
     * Step the model for any other sample times *
     *********************************************/
        tid += FIRST_TID;
        MdlOutputs(tid);
#ifdef EXT_MODE	  
        /* Buffer signal point according to the Discarding Factor (disFac) */
        if (++counter[tid] >= disFac) {
            counter[tid] = 0;
            /* If disFac = 0 all the packets are to be discarded. */
            if (disFac)
                    rtExtModeUpload(tid, rtmGetTaskTime(S,tid));
        }  
#endif		
        MdlUpdate(tid);
        rt_SimUpdateDiscreteTaskTime(rtmGetTPtr(S), 
                                         rtmGetTimingData(S),tid);
    }

} /* end rtOneStep */

#endif /* MULTITASKING */


/*===================*
 * Visible functions *
 *===================*/


/* NOTE:
 * The following three functions: model_init, model_run and model_clean, were before        *
 * grouped in main(). This function are broken up because they are called separately from   *
 * krnl_main.c           */

/* Function: model_init  ============================================================
 *
 * Abstract:
 *      initialises  model 
 */
int_T model_init(RT_MODEL *S)
{
    const char *status;

    /******************************
     * MathError Handling for BC++ *
     ******************************/
#ifdef BORLAND
    signal(SIGFPE, (fptr)divideByZero);
#endif


    /****************************
     * Initialize global memory *
     ****************************/
    (void)memset(&GBLbuf, 0, sizeof(GBLbuf));

    /************************
     * Initialize the model *
     ************************/
    if(rt_InitInfAndNaN(sizeof(real_T))) 
		return ERROR;
     
    if (FT >= 0.0 || FT == RUN_FOREVER) rtmSetTFinal(S,FT);

    MdlInitializeSizes();
    MdlInitializeSampleTimes();
    
    status = rt_SimInitTimingEngine(rtmGetNumSampleTimes(S),
                                    rtmGetStepSize(S),
                                    rtmGetSampleTimePtr(S),
                                    rtmGetOffsetTimePtr(S),
                                    rtmGetSampleHitPtr(S),
                                    rtmGetSampleTimeTaskIDPtr(S),
                                    rtmGetTStart(S),
                                    &rtmGetSimTimeStep(S),
                                    &rtmGetTimingData(S));

    if (status != NULL) {
        PRINT_W("Failed to initialize sample time engine: %s\n", status);
        return ERROR;
    }
    rt_CreateIntegrationData(S);

    GBLbuf.errmsg = rt_StartDataLogging(rtmGetRTWLogInfo(S),
                                        rtmGetTFinal(S),
                                        rtmGetStepSize(S),
                                        &rtmGetErrorStatus(S));
    if (GBLbuf.errmsg != NULL) {
        PRINT_W("Error starting data logging: %s\n",GBLbuf.errmsg);
        return(EXIT_FAILURE);
    }
    dataLogInit = TRUE;

    PRINT_M("\n** starting the model **\n");

    MdlStart();
    if (rtmGetErrorStatus(S) != NULL) {
      GBLbuf.stopExecutionFlag = 1;
    }
    
    return SUCCESS;    
    
} /* End of model_init */


/* Function: model_run =============================================================
 *
 * Abstract:
 *      run model 
 */

void model_run(RT_MODEL *S, int_T tid) 
{
    static int stopThreads = FALSE;

    if (!GBLbuf.stopExecutionFlag && (rtmGetTFinal(S) == RUN_FOREVER ||
                         rtmGetTFinal(S)-rtmGetT(S) > rtmGetT(S)*DBL_EPSILON)) {
        rt_OneStep(S, tid);

    } else {
        /* Execute model last time step */ 	
        if (!GBLbuf.stopExecutionFlag && !rtmGetStopRequested(S)) {
            rt_OneStep(S, tid);
        }
        /* Do stop_model just one (remeber multitasking) */		
        if (stopThreads == FALSE) {
            stopThreads = TRUE;
            /* Cancel the threads and terminate the application. */
            stop_model();
        }
    }
}  /* end model_run */


/* Function: model_clean =============================================================
 *
 * Abstract:
 *      clean up and exit the model 
 */
int_T model_clean(RT_MODEL *S)
{
    /********************
     * Cleanup and exit *
     ********************/
    if (dataLogInit == TRUE)
        rt_StopDataLogging(rtmGetRTWLogInfo(S));  

#ifdef EXT_MODE
    UploadPrepareForFinalFlush();  /* This function used to be at the begining of rt_ExtModeShutDown() */
    rt_UploadServerWork_rtl();    /* now in ext_svr_rtl() */
  
    /* rt_ExtModeShutdown();   NOTE: this function is called in us_main.c
     * NOTE:
     * UploadShutDown() was originally in ext_svr.c / rt_ExtModeShutdown()*/

    UploadShutDown(); 
#endif

    if (GBLbuf.errmsg) {
        PRINT_W("%s\n",GBLbuf.errmsg);
        return ERROR;
    }

    if (rtmGetErrorStatus(S) != NULL) {
        PRINT_W("%s\n", rtmGetErrorStatus(S));
        return ERROR;
    }
    MdlTerminate();
    return(EXIT_SUCCESS);

} /* end main */



/* EOF: grt_main.c */
