/********************************************************************
 * Simulink Target for Real-Time Linux (ST-RTL)                     *
 *                                                                  *
 * Author:       Raul Murillo Garcia <r.murillogarcia@gcal.ac.uk>   *
 * Last review:  30 August 2002                                     *
 *                                                                  *
 * Driver file for the data acquistion card CIO-DAS08/JR-AO,        *
 * of Measurement Computing.                                        *
 *                                                                  *
 * This program is free software; you can redistribute it and/or    *
 * modify it under the terms of the GNU General Public License      *
 *                                                                  *
 ********************************************************************/
 
#include <asm/io.h>
#include <linux/ioport.h>

/* rtl_printf() */
#include <rtl_printf.h>
#include <pthread.h>

#include "tmwtypes.h" /* for real_T */
#include "common.h"
#include "ciodas08jrao.h" 

#define printf rtl_printf

/********************
 * GLOBAL VARIABLES *
 ********************/
 
/* Possible base addresses. */
static unsigned int dacPorts[] = {0x200, 0x270, 0x2e0, 0x2e8, 0x2f0,   
                                  0x2f8, 0x300, 0x310, 0x320, 0x378, 0};
/* Base address. */
int base;
					    
/* Mutex to protect channel reading/writing. Note that multiple thread 
 * can access the driver which would lead to scrambled and incoherent results. 
 * Mutex control the access to the API reading/writing functions.
 */
pthread_mutex_t m_read, m_write;

					    
/**************************
 **** PRIVATE FUNCTIONS ***
 **************************/

/* FUNCTION: read_digital(...)
 * Read from digital inputs    
 * 
 * mask:   masks out the channel
 */
static void read_digital(int channel, real_T *value)
{
  unsigned short int value_r;
  unsigned char mask;
  
  /* on the protection board the digital input channels starts at 9, on the DAC they start at 0 */
  channel -= 9;
  
  /* mask has a bit set at position "channel" (0..7) */
  mask = 0x01 << channel;
  
  value_r = (inb_p(base + DIGITAL_IO));
  value_r =  (value_r & mask);

  /* value can only be either 1 or 0 */
  if (value_r) *value = 1.0F;
  else *value = 0.0F;
  
} /* end read_digital(...) */

/* FUNCTION: write_digital(...)
 * Write to digital outputs    
 * 
 * old_value:  stores previous value of the 8 digital channels in order not to overwrite them
 * value_w:    is the unsigned short int representation of a floating point, whose value
 *             can only be either 1 or 0
 * mask:       masks old_value to target a specific bit (channel) of the register
 */
static void write_digital(int channel, real_T value)
{
  unsigned short int value_w, old_value;
  unsigned char mask;
  
  /* On the protection board the digital output channels start at 3, on the DAC they start at 0 */
  channel -= 3;
  
  /* In order not to overwrte the rest of channels the state of each of them must be known*/
  old_value = inb_p(base + DIGITAL_IO);
  
  /* Mask has a bit set at position "channel" (0..7) */
  mask = 0x01 << channel;
  
  /* If channel has to be set, the current value of the 8 digital channels is ORed    *
   * with mask, whose value has all zeros but a one at the bit position "channel"     *
   * Value can only be 1 or 0. If different than 0, value is 1                        *
   */
  
  if (value!=0.0F) {
    value_w = 1;
    value_w = mask | old_value;
  }
  
  /* If channel has to be reset, the current value of the 8 digital channels is ANDed  *
   * with mask, whose value has all ones but a zero at the bit position "channel"      *
   */
  
  else {
    value_w = 0;
    mask ^= 0xFF; /* invert bits to AND */
    value_w = mask & old_value;
  }
  
  outb_p(value_w,base + DIGITAL_IO);
  
} /* End write_digital(...) */

/* FUNCTION: read_analog(...)
 * Read from analog inputs    
 * 
 * lsb: 4 lest significant bits of the 12 bit value 
 * msb: 8 most significant bits of the 12 bit value
 * value_r is normalised to -1  +1
 *
 */

static void read_analog(int channel, real_T *value)
{
  unsigned char lsb, msb;
  unsigned short int value_r;
  
  /* on the protection board the analog input channels starts at 1, on the DAC they start at 0 */
  channel --;

  /* select channel */
  outb_p( channel, base + ANALOG_IN_CHANNEL);
  
  /* start conversion */
  outb_p( 0xFF, base + ANALOG_IN_START);

  /* Wait for conversion to finish. If BLCK_WAIT is defined use usleep().
   * Note that usleep() might take considerably longer than the specified AD_DELAY. For quicker 
   * respones don't define BLCK_WAIT (user parameters through SImulink inerface). 
   * The drawback is that the CPU is kept busy polling the ready bit.
   */
#ifndef BLCK_WAIT
   usleep(AD_DELAY);    /* Non-blocking function. */
#endif 

  /* Wait for ready bit (EOC clear)*/
   while( 0x80 & inb_p( base + ANALOG_IN_CRTL ) ) ;
 
  /* An analog value has 12 bits. The 4 lest significant bits are supplied by the        *
   * 4 msb of the "base" (ANALOG_IN_LSB). The 8 most significant bits are supplied   *
   * by "base" + 1 (ANALOG_IN_MSB)                                                   */
  lsb = inb_p(base + ANALOG_IN_LSB) & 0xF0;  
  msb = inb_p(base + ANALOG_IN_MSB);
  /* combine lsb and msb to have a single 12 bit value */
  value_r = (((int)lsb) >> 4) + (((int)msb) << 4);
  
  /* normalise value */
  *value = ((float) (((float)(value_r *2))/ 0xFFF) - 1);
  
  /* the value must be within the range + 1 -1 */
  if ( *value < -1.0F ) *value = -1.0F;
  else if ( *value > 1.0F ) *value = 1.0F;

} /* End read_analog(...) */

/* FUNCTION: write_analog(...)
 * Write to analog outputs    
 * 
 * There are two 12 bit analog outputs,that can written by accessing the ANALOG_OUT_LSB  
 * and ANALOG_OUT_MSB. To start the DA conversion DIGITAL_IO has to be read.
 * value_w is normalised to -1  +1
 */
static void write_analog(int channel, real_T value)
{
  unsigned short int value_w;
  
  /* On the protection board the analog ouput channels start at 1, on the DAB they start at 0 */
  channel --;
  
  /* Value must be wtihing the range -1 + 1 */ 
  if ( value < -1.0F ) value = -1.0F;
  else if ( value > 1.0F ) value = 1.0F;
  
  /* Normalise value */
  value_w = (unsigned short int) (((value + 1)* 0xFFF) / 2);
  
  if (channel==0) {
    outb_p((unsigned char)((value_w >> 8) & 0x0F), base + ANALOG_OUT_MSB0 );
    outb_p((unsigned char)(value_w & 0xFF), base + ANALOG_OUT_LSB0 );
  }
  else if(channel == 1) {
    outb_p((unsigned char)((value_w >> 8) & 0x0F), base + ANALOG_OUT_MSB1 );
    outb_p((unsigned char)(value_w & 0x00FF), base + ANALOG_OUT_LSB1 );
  }
  /* read Digital I/O control register to initiate DAC */
  inb_p(base + DIGITAL_IO);
  
}  /* End write_analog(...) */


/**************************
 **** PUBLIC FUNCTIONS ****
 **************************/


/* FUNCTION: as08jrao_write(...)
 * Public funtion to write outputs    
 * 
 * This is the API function used to write a value to a channel (either analog or digital) 
 */
extern void das08jrao_write(int channel, real_T value)
{
  /* Lock writing mutex. */
  pthread_mutex_lock(&m_write);

  /* There are a total of 10 output channels, 2 analogs and 8 digitals.   */
  /* Check that the channel is within the range 1 - 10 */
  if ( (channel > 1) || (channel < 10)) {
    if (channel > 2) {
      /* digital output */
      write_digital(channel, value);
    }
    else {
      /* analog output  */
      write_analog(channel, value);
    }
  }
  
  /* Unlock writing mutex. */
  pthread_mutex_unlock(&m_write);  
  
}  /* End of das08jrao_write(...) */


/* FUNCTION: as08jrao_read(...)
 * Public funtion to read inputs    
 * 
 * This is the API function used to read a value from a channel (either analog or digital) 
 */
extern void das08jrao_read(int channel, real_T *value)
{
  /* Lock reading mutex. */
  pthread_mutex_lock(&m_read);

  /* There are a total of 16 input channels, 8 analogs and 8 digitals.   */
  /* Check that the channel is within the range 1 - 16 */
  if ( (channel > 1) || (channel < 16)) {
    if (channel > 8) {
      /* digital input */
      read_digital(channel, value);
    }
    else {
      /* analog input  */
      read_analog(channel, value);
    }
  }
  
  /* Unlock reading mutex. */
  pthread_mutex_unlock(&m_read);
  
} /* End of das08jrao_read(...) */


/* FUNCTION: das08jrao_init()
 * Public funtion to detect the CIO-DAS08/JR-AO card.
 *  
 */
extern int das08jrao_init()
{
  int i = 0;    

  while ((base = dacPorts[i++]) != 0) {
    /* Probe port. */
    if (-EBUSY == check_region(base, DAS08JRAO_SIZE)) {
      continue;
    }

    request_region(base, DAS08JRAO_SIZE, DAS08JRAO_NAME);

    /* Try bit 8 of CONTROL port. It should be zero because there
     * is no A/D conversion at the moment. */
    if ( 0x80 == ( 0x80 & inb_p(base + ANALOG_IN_CRTL))) {
      release_region(base, DAS08JRAO_SIZE);
      continue;
    }
    
    /* Try the port. 
     * Set the control address to read analog input 3. */
    outb_p(0xFF, base + ANALOG_IN_CRTL);
    outb_p(0x03, base + ANALOG_IN_CRTL);
    if ( 0x03 != ( 0x07 & inb_p(base + ANALOG_IN_CRTL))) {
      release_region(base, DAS08JRAO_SIZE);
      continue;
    }

    /* Do the same with channel 6. */
    outb_p(0x06, base + ANALOG_IN_CRTL);
    if ( 0x06 != ( 0x07 & inb_p(base + ANALOG_IN_CRTL))) {
      release_region(base, DAS08JRAO_SIZE);
      continue;
    }

    break;
  }
  
  if (base == 0) {
    return -ENODEV;
  }
  else {
    PRINT_M("\nData Acqusition Card CIO-DAS08/JR-AO found on 0x%lx \n\n", base);
  }
  /* Create mutex */
  pthread_mutex_init(&m_read, NULL);
  pthread_mutex_init(&m_write, NULL);
  
  return TRUE;
  
} /* End of das08jrao_init(...) */

/* FUNCTION: das08jrao_close()
 * Public funtion to close the CIO-DAS08/JR-AO card.
 *  
 */
extern void das08jrao_close()
{
  if (base != 0)  
   release_region(base, DAS08JRAO_SIZE);
  
  /* Destroy mutex */
  pthread_mutex_destroy(&m_read);
  pthread_mutex_destroy(&m_write);
  
} /* End of das08jrao_close() */