/*
 * Copyright 1994-2002 The MathWorks, Inc.
 *
 * File: ext_svr.c     $Revision: 1.51 $
 *
 * Abstract:
 *  Provides the following functions:
 *  ProcessConnectMsg_rtl(), processes the checksum.
 *  rt_UploadServerWork_rtl(), writes a signal point to shared memory, from where 
                 it is later read in ext_svr_us.c and transmitted to the host machine4. 
 *  df_update(), discarding algorithm to update the discarding factor. Used when the macro
 *               RT_MONITORING is TRUE to guarantee real-time monitoring in a LAN.
 *  Functions for allocating, initializating and freeing the structure SendData.
 *
 */ 
 
 
/*****************
 * Include files *
 *****************/
 
/* Shared Memory */
#include "mbuff.h"

/* ST-RTL header */
#include "common.h"
#include "ext_svr_rtl.h"

/* Real Time Workshop headers*/
#include "tmwtypes.h" 
#include "simstruc.h"
#include "ext_share.h"
#include "updown.h"
#include "updown_util.h"
#include "dt_info.h"

/* rtl_printf() */
#include <rtl_printf.h>
#define printf rtl_printf

volatile SendDataStruc *SendData;
volatile uint32_T      *tmpBuf;


int   nBytes      = 0; /* Queue size in bytes.           */
int   queueSize   = 0; /* Current queue size in 100%.    */
int   errAvg      = 0; /* Average error. The low-pass filter is an Exponential Weighted Moving Average. */
int   cP          = 0; /* Proportional part of the controller. */
float cI          = 0; /* Integral part of the controller.     */
int   output1     = 1; /* Output from the # of SP the host reads per second. */ 
int   output2     = 1; /* Output of the PI controller.      */
int   disFac      = 1; /* The current discarding factor. */
int   oneSecTimer = DA_RATE; /* This variable is decremented from DA_RATE to 0 to count 1 second. */


/**********************
 * External Variables *
 **********************/
 
/* nBytesSec: number of bytes per second to be sent to the Simulink host. This value is calculated 
              in updown.c and used here to work out the buffer size in SendData.                 */
extern volatile int_T nBytesSec;  
/* nSPsec: number of signal points per second to be sent to the Simulink host. This value is calculated 
              in updown.c and used here to work out the minimum discarding factor.                    */
extern volatile int_T nSPsec;


/* Function ====================================================================
 * Sends data stored in rtl to ext_svr that it can be send to the host.
 */
int_T ProcessConnectMsg_rtl(RTWExtModeInfo *ei)
{
	int_T                   tmpBufSize;
  
    const DataTypeTransInfo *dtInfo    = rteiGetModelMappingInfo(ei);
    uint_T                  *dtSizes   = dtGetDataTypeSizes(dtInfo);
    int_T                   nDataTypes = dtGetNumDataTypes(dtInfo);
  
	/* Send 2nd EXT_CONNECT_RESPONSE message containing the following 
	 * fields:
	 *
	 * CS1 - checksum 1 (uint32_T)
	 * CS2 - checksum 2 (uint32_T)
	 * CS3 - checksum 3 (uint32_T)
	 * CS4 - checksum 4 (uint32_T)
	 * 
 	 * targetStatus  - the status of the target (uint32_T)
	 *
	 * nDataTypes    - # of data types        (uint32_T)
	 * dataTypeSizes - 1 per nDataTypes       (uint32_T[])
	 */
  
	int nMsgEls    =  4 +                        /* checkSums       */
				1 +                        /* targetStatus    */
				1 +                        /* nDataTypes      */
				dtGetNumDataTypes(dtInfo); /* data type sizes */
  
	tmpBufSize = nMsgEls * sizeof(uint32_T);

	/* tmpBuf is allocated in shared memory so that ext_svr can access it */
	tmpBuf=(uint32_T *) mbuff_alloc("tmpBuf",tmpBufSize);
	if (tmpBuf == NULL)  
		return ERROR;
	
	/* Checksums, target status & SL_DOUBLESize. */
	tmpBuf[0] = rteiGetChecksum0(ei);
	tmpBuf[1] = rteiGetChecksum1(ei);
	tmpBuf[2] = rteiGetChecksum2(ei);
	tmpBuf[3] = rteiGetChecksum3(ei);
 	/* tmpBuf[4] is filled in ext_svr.c   */
	/* nDataTypes and dataTypeSizes */
	tmpBuf[5] = (uint32_T)nDataTypes;
	(void)memcpy((uint32_T *) &tmpBuf[6], dtSizes, sizeof(uint32_T) * nDataTypes);
  
	return  tmpBufSize;
  
} /* end ProcessConnectMsg_rtl */


/* Function: rt_UploadServerWork_rtl ===============================================
 * Upload model signals to the user space through FIFO.
 */
void rt_UploadServerWork_rtl(void)
{
	ExtBufMemList upList;
	unsigned int nBytesLeft, nBytesToAdd, i;

	if (SendData->status == CONNECT || SendData->status == HOLD) {
		if (SendData->status == HOLD) {
			UploadCancelLogging();
			SendData->status = CANCEL;
		}
		UploadBufGetData(&upList);
		while(upList.nActiveBufs > 0) {
			for (i=0; i<upList.nActiveBufs; i++) {
				const BufMem *bufMem = &upList.bufs[i];
	
				if (SendData->head == SendData->tail) 
					nBytesLeft = SendData->size;
				else 
					nBytesLeft = (SendData->tail + SendData->size - SendData->head) % SendData->size;
				
				nBytesToAdd = bufMem->nBytes1 + bufMem->nBytes2;

			 	/* Leave 8 bytes for the TRIGGER_TERMINATING message; and 1 byte *
				 * so that head and tail can never be the same when the buffer   *
				 * is full. Otherwise, when emptying the buffer, the full state  *
				 * is seen as empty sinve head==tail.                            */
				if (nBytesLeft < (nBytesToAdd + 9) && nBytesToAdd != 8) {
					PRINT_W("\n!!! The buffer is full. Decrease the sample rate !!!\n");
					UploadCancelLogging();
				} else  {
					/* Check whether head has reached the end of the buffer. 
					 * If so mark the byte with the '0000' marker.
				 	 */
					if ((SendData->head + nBytesToAdd + 4) > SendData->size) {
						/* Set two bytes to zero ('0000' marker) */
						memset(SendData->rt_buff + SendData->head, 0, 4);
						SendData->head = 0;
					}
					
					/* Copy Simulink Packet onto shared memory */
					(void) memcpy((char_T *) (SendData->rt_buff + SendData->head), \
										bufMem->section1, bufMem->nBytes1);
					if (bufMem->section2 != NULL) 
						(void) memcpy((char_T *) (SendData->rt_buff + SendData->head), \
											bufMem->section2, bufMem->nBytes2);

					SendData->head += nBytesToAdd;
				}
				/* Comfirm that the data was sent */
				UploadBufDataSent(upList.tids[i]);
			}
			UploadBufGetData(&upList);
		} 
	}
} /* end rt_UploadServerWork */

#ifdef RT_MONITORING
/* Function: df_update ===============================================
 * Runs the Discarding Algorithm and updates the Discarding Factor.  
 */
void df_update()
{
    if (SendData->status == CONNECT) {
		/* Decrease the timer. */
		if (--oneSecTimer == 0) {
			/* Request host net accesses */
			SendData->hostNetRate = 1;
			/* Reset the timer with DA_RATE, the Discarding Algorithm rate. */
			oneSecTimer = DA_RATE;
		}
		/* Calculate the disFacMin for the new value of host network reading rate. */
		if (SendData->hostNetRate > 1) {
			int temp;
			temp = (nSPsec) / (SendData->hostNetRate + 1) + 1;
			if (output1 == 1) { /* First time */
				/* After one second the host machine cannot read all the signal points sent, since some are  *
				* still travelling or in the host network input queue. That is why even when the network-   *
				* -hardware can handle all the data to be transmitted the value of 'temp' will be just      *
				* above 1. Since 'temp' is round up the final value is 2. We avoid this by cheking after    *
				*  the first second that the system is stable, i. e. errAvg = 0 with minimum disFac.        */
				if (disFac == 1 && errAvg ==0)
					temp = 1;
				else
					output1 = temp; 
			}
			else if (output1 > temp)
				output1 --; /* Slow decrease, not abrupt */

			SendData->hostNetRate = 0; /* Avoid reentering the this 'if' before the timer expires again. */ 
		}

		/* Check how many bytes are buffered */
		nBytes = SendData->head - SendData->tail;
		if (nBytes < 0 ) 
			nBytes = (SendData->size - SendData->tail) + SendData->head;

		queueSize = (100 * nBytes) / (SendData->size); /* Queue size in % */

		/* If the buffer occupancy is above MAX_QUEUE_SIZE all packets are discarded */
		if (queueSize < 100) {
			/* Filter the errror. The Low Pass Filter is an exponential weighted moving average  *
			* We shift instead of multiply because it is more effective. Shifting 5 bits yields *
			* a factor of 2^(-5) = 0.03125.                                                     *
			* errAvg(n) = errAvg(n-1) * (1 - Wq) + queueSize(n) * Wq, where Wq = 0.03125        */
			errAvg += (queueSize - errAvg) >> 5; 

			/* When the system is stable (empty buffer) there is no need for I controller. We reset it. */
			if (queueSize == 0) 
				cI = 0;
	   
			/* PI control: y(n) = x(n)*P_GAIN + (Sum(cI(0) to cI(n-1)) + x(n)*I_GAIN) */
			cP      = errAvg << 1;             /* Proportional actoin of the controller. GAIN = 2.  */
			output2 = cP + (int) cI;           /* Output of the PI controller.                      */
			cI     += ((float)errAvg * 0.005); /* Integral action of the controller. GAIN  = 0.005. */

			if (output2 >= output1)
				disFac = output2;
			else 
				disFac = output1;
		} else 
			disFac = 0;
	}

} /* end df_udpate  */
#endif

/* Function: SendDataInit ===============================================
 * Initializes strucutre SendData 
 */
void  SendDataInit(void)
{
	SendData->status      = CONNECT;
	SendData->head        = 0;
	SendData->tail        = 0;
	SendData->hostNetRate = 0;
} /*end SendDataInit */


/* Function: SendDataBuffAlloc ===============================================
 * Allocates shared memory for buffer in SendData. SendData->USbuf is allocated 
 * according to the number of bytes to be sent every cycle. Therefore whenever 
 * a disconnect/connect or cancel/arm occurs the buffer has to be  freed before 
 * reallocaing it.
 */
int_T  SendDataBuffAlloc(void)
{
	SendData->size = nBytesSec/10; /* Data produced in one second */
	SendData->rt_buff = (char_T *) mbuff_alloc("buffer", SendData->size);
	if (SendData->rt_buff == NULL || SendData->size == 0) 
		return ERROR;

	return SUCCESS;
	
} /*end SendDataBuffAlloc */


/* Function: SendDataAlloc ===============================================
 * Allocates shared memory for SendData 
 */
int_T SendDataAlloc(void)
{
	SendData = (volatile SendDataStruc *) mbuff_alloc("SendData", sizeof(SendDataStruc));
	if (SendData == NULL) 
		return ERROR;
	
	SendData->status = DISCONNECT;
	SendData->rt_buff = NULL;
  
	return SUCCESS;
	
} /*end SendDataAlloc */


/* Function: SendDataBuffFree ===============================================
 * Frees shared memory for buffer in SendData 
 */ 
void   SendDataBuffFree(void)
{
	if (SendData->rt_buff != NULL) 
		mbuff_free("buffer", SendData->rt_buff);
} /*end SendDataBuffFree*/


/* Function: SendDataFree ===============================================
 * Frees shared memory for SendData 
 */ 
void   SendDataFree(void)
{
	SendDataBuffFree();  /* Free the las allocation of SendData->rt_buff */
	mbuff_free("SendData", (void *) SendData);

} /*end SendDataFree*/

/* Function: tmpBufFree ===============================================
 * Frees shared memory for tmpBuf
 */
void   tmpBufFree(void)
{
	mbuff_free("tmpBuf", (void *) tmpBuf);

} /*end tmpBufFree*/