/*
 * Copyright 1994-2002 The MathWorks, Inc.
 *
 * File: ext_svr.c     $Revision: 1.51 $
 *
 * Abstract:
 *  External mode server interface (TCPIP example).  Provides functions
 *  that get called by main routine (modelname.c):
 *    o ExtParseArgsAndInitUD:  parse args and create UserData
 *    o ExtWaitForStartMsg:     return true if waiting for host to start
 *    o rt_ExtModeInit:         external mode initialization
 *    o rt_ExtModeSleep:        pause the process
 *    o rt_MsgServerWork:       server for setting/getting messages from host
 *    o rt_MsgServer:           server dispatcher - for multi-tasking targets
 *    o rt_UploadServerWork:    server for setting upload data on host
 *    o rt_UploadServer:        server dispatcher - for multi-tasking targets
 *    o rt_ExtModeShutdown:     external mode termination
 *
 *  Paremter downloading and data uploading supported for single and
 *  multi-tasking targets.
 */

/*****************
 * Include files *
 *****************/

/*ANSI C headers*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

/*Real Time Workshop headers*/
#include "tmwtypes.h"
#include "simstruc_types.h"

#include "ext_share.h"
#include "ext_test.h"
#include "ext_svr_transport.h"
#include "updown_util.h"
#include "dt_info.h"

/* ST-RTL header */
#include "common.h"

/* Shared Memory header */
#include "mbuff.h" 

/*Uncomment to test 4 byte reals*/
/*#define real_T float*/

#ifndef __LCC__
#define UNUSED_PARAM(p) (void)((p))
#else
#define UNUSED_PARAM(p) /* nothing */
#endif


/**********************
 * External Variables *
 **********************/
#if (WAIT_START_MODEL == 1) 
/* Wait for START message from Simulink side.Three conditions are needed to create the
 * RT thread. Host-Target need to be connected, trigger armed and model started. When one
 * of this conditions is true we set to high bit 0, 1 or 2 of startModel. Therefore, when the 
 * value is ...0000111 the thread can be staretd. On the other hand when a disconnection occurs
 * or the trigger is cancelled we set to low the corresponding bit.
 * CONNECT: bit 0
 * START:   bit 1
 * ARM:     bit 2
 */ 
extern volatile int_T  startModel;
TargetSimStatus modelStatus = TARGET_STATUS_WAITING_TO_START;
#else
TargetSimStatus modelStatus = TARGET_STATUS_RUNNING;
#endif

extern int_T fd1;  /* file descriptors for the FIFO */


/********************
 * Global Variables *
 ********************/

/*
 * Flags.
 */
PRIVATE boolean_T   connected       = FALSE;
PRIVATE boolean_T   commInitialized = FALSE;

/*
 * Pointer to opaque user data (defined by ext_svr_transport.c).
 */
PRIVATE ExtUserData *extUD          = NULL;

/*
 * Buffer used to receive messages.
 */
PRIVATE int_T msgBufSize = 0;
PRIVATE char  *msgBuf    = NULL;

PRIVATE volatile SendDataStruc *SendData = NULL;  
PRIVATE volatile char          *msg_shm  = NULL; 

/*******************
 * Local Functions *
 *******************/

/* Function: SendDataBuffAlloc ===============================================
 * Allocates shared memory for buffer in SendData. SendData->net_buff is allocated 
 * according to the number of bytes to be sent every cycle. Therefore whenever 
 * a disconnect/connect or cancel/arm occurs the buffer has to be  freed before 
 * reallocaing it.
 */
PUBLIC int_T  SendDataBuffAlloc(void)
{

	SendData->net_buff = (char_T *) mbuff_alloc("buffer", SendData->size);
	if (SendData->net_buff == NULL) {
		PRINT_E(stderr,"\nError: Shared memory for buffer could not be allocated.");
		return ERROR;
	}
} /*end SendDataBuffAlloc */


/* Function: SendDataAlloc ===============================================
 * Allocates shared memory for SendData 
 */
PUBLIC int_T  SendDataAlloc(void)
{
	SendData = (SendDataStruc *) mbuff_alloc("SendData", sizeof(SendDataStruc));
	if (SendData == NULL) {
		PRINT_E(stderr,"\nError: Shared memory for structure SendData could not be allocated"
			"\n\tDecrease the value of MEMORY in common.h\n");
		return ERROR;
	}
	SendData->net_buff = NULL;
	
	return SUCCESS;
} /*end SendDataAlloc */


/* Function: SendDataBuffFree ===============================================
 * Frees shared memory for buffer in SendData 
 */ 
PUBLIC void  SendDataBuffFree(void)
{
	/* If SendData == NULL RTLbuff is not allocated */
	if (SendData->net_buff != NULL)       
		mbuff_free("buffer", (void *) SendData->net_buff);
	
} /*end SendDataBuffFree*/


/* Function: SendDataFree ===============================================
 * Frees shared memory for SendData 
 */
PUBLIC  void  SendDataFree(void)
{
	if(SendData != NULL) {
		/* Free buffer in user space. */
		SendDataBuffFree();  	
		mbuff_free("SendData", (void *) SendData);
	}
} /* end SendDataFree */


/* Function: msgFree ===============================================
 * Frees shared memory for msg_shm 
 */
PUBLIC void msgFree(void)
{
	mbuff_detach("msg_shm", msg_shm);
} /* end msgFree* /


/* Function: GrowRecvBufIfNeeded ===============================================
 * Abstract:
 *  Allocate or increase the size of buffer for receiving messages from target.
 */
PRIVATE boolean_T GrowRecvBufIfNeeded(const int msgSize)
{
    if (msgSize > msgBufSize) {
        if (msgBuf != NULL) free(msgBuf);

        msgBufSize = msgSize;
        msgBuf     = (char *)malloc(msgBufSize);

        if (msgBuf == NULL) return(EXT_ERROR);
    }
    return(EXT_NO_ERROR);
} /* end GrowRecvBufIfNeeded */


/* Function: GetMsgHdr =========================================================
 * Abstract:
 *  Attempts to retrieve a message header from the host.  If a header is in 
 *  fact retrieved, the reference arg, 'hdrAvail' will be returned as true.
 *
 *  EXT_NO_ERROR is returned on success, EXT_ERROR is returned on failure.
 *
 * NOTES:
 *  o It is not necessarily an error for 'hdrAvail' to be returned as false.
 *    It typically means that we were polling for messages and none were
 *    available.
 */
PRIVATE boolean_T GetMsgHdr(MsgHeader *msgHdr, boolean_T *hdrAvail)
{
    int_T     nGot  = 0; /* assume */
    boolean_T error = EXT_NO_ERROR;
    
    /*
     * Read the message header.  Assumed that the entire message
     * header can be read in one shot.
     */
    error = ExtGetHostMsg(extUD,sizeof(MsgHeader),&nGot,(char_T *)msgHdr);
    if (error) goto EXIT_POINT;
    assert((nGot == 0) || (nGot == sizeof(MsgHeader)));

EXIT_POINT:
    *hdrAvail = (boolean_T)(nGot > 0);
    return(error);
} /* end GetMsgHdr */


/* Function: GetMsg ============================================================
 * Abstract:
 *  Receive nBytes from the host.  Return a buffer containing the bytes or
 *  NULL if an error occurs.  Note that the pointer returned is that of the
 *  global msgBuf.  If the buf needs to be grown to accommodate the message,
 *  it is realloc'd.  This function will try to get the requested number
 *  of bytes indefinately - it is assumed that the data is either already there,
 *  or will show up in a "reasonable" amount of time.
 */
PRIVATE const char *GetMsg(const int msgSize)
{
    int_T     nGot;
    boolean_T error     = EXT_NO_ERROR;
    int_T     nGotTotal = 0;

    error = GrowRecvBufIfNeeded(msgSize);
    if (error != EXT_NO_ERROR) goto EXIT_POINT;
    
    /* Get the data. */
    while(nGotTotal < msgSize) {
        error = ExtGetHostMsg(extUD,
            msgSize - nGotTotal, &nGot, (char_T *)(msgBuf + nGotTotal));
        if (error) {
	    PRINT_E(stderr,"ExtGetHostMsg() failed.\n");
            goto EXIT_POINT;
	}

	nGotTotal += nGot;
    }

EXIT_POINT:
    return((error == EXT_NO_ERROR) ? msgBuf : NULL);
} /* end GetMsg */


/* Forward declaration */
boolean_T rt_UploadServerWork(void);


/* Function: DisconnectFromHost ================================================
 * Abstract:
 *  Disconnect from the host.
 */
PRIVATE void DisconnectFromHost(void)
{
/*    UploadPrepareForFinalFlush(); */
    rt_UploadServerWork();
    
/*    UploadLogInfoTerm(); */

    connected       = FALSE;
    commInitialized = FALSE;

    ExtCloseConnection(extUD);
} /* end DisconnectFromHost */


/* Function: ProcessConnectMsg =================================================
 * Abstract:
 *  Process the EXT_CONNECT message and send response to host.
 */
PUBLIC boolean_T ProcessConnectMsg(int_T tmpBufSize)
{
    int_T                   nSet;
    MsgHeader               msgHdr;
    uint32_T                *tmpBuf = NULL;
    boolean_T               error   = EXT_NO_ERROR;
    
    assert(connected);
    assert(!comminitialized);

    /*
     * Send the 1st of two EXT_CONNECT_RESPONSE messages to the host. 
     * The message consists purely of the msgHeader.  In this special
     * case the msgSize actually contains the number of bits per byte
     * (not always 8 - see TI compiler for C30 and C40).
     */
    msgHdr.type = (uint32_T)EXT_CONNECT_RESPONSE;
    msgHdr.size = (uint32_T)8; /* 8 bits per byte */

    error = ExtSetHostMsg(extUD,sizeof(msgHdr),(char_T *)&msgHdr,&nSet);
    if (error || (nSet != sizeof(msgHdr))) {
        PRINT_E(stderr,
            "ExtSetHostMsg() failed for 1st EXT_CONNECT_RESPONSE.\n");
        goto EXIT_POINT;
    }

	/* Send 2nd EXT_CONNECT_RESPONSE message containing the following 
	 * fields:
	 *
	 * CS1 - checksum 1 (uint32_T)
	 * CS2 - checksum 2 (uint32_T)
	 * CS3 - checksum 3 (uint32_T)
	 * CS4 - checksum 4 (uint32_T)
	 * 
	 * targetStatus  - the status of the target (uint32_T)
	 *
	 * nDataTypes    - # of data types        (uint32_T)
	 * dataTypeSizes - 1 per nDataTypes       (uint32_T[])
	 * the data follow the following pattern 
	 *        4  bytes for checkSums       
	 *        1  bytes for targetStatus    
	 *        1  byte  for nDataTypes     
	 *        ?  bytes for  data type sizes 
	 * 
	 * the first value of SendData is the number of bytes sent by ext_svr_rtl 
	 *
	 * use shared memory, where data to be sent to the target is stored 
	 */
    tmpBuf=(uint32_T *) mbuff_attach("tmpBuf", tmpBufSize);
    if (tmpBuf == NULL) {
        PRINT_E(stderr, "\n!!!! Shared memory could not be allocat");
        goto EXIT_POINT;
    }

    tmpBuf[4] = (uint32_T)modelStatus;  /* this could not be filled in ext_svr_rtl */
   
    /* Send message header. */
    msgHdr.type = EXT_CONNECT_RESPONSE;
    msgHdr.size = tmpBufSize;

    error = ExtSetHostMsg(extUD,sizeof(msgHdr),(char_T *)&msgHdr,&nSet);
    if (error || (nSet != sizeof(msgHdr))) {
        PRINT_E(stderr,
            "ExtSetHostMsg() failed for 2nd EXT_CONNECT_RESPONSE.\n");
        goto EXIT_POINT;
    }
   
    /* Send the message. */
    error = ExtSetHostMsg(extUD,tmpBufSize,(char_T *)tmpBuf,&nSet);
    if (error || (nSet != tmpBufSize)) {
        PRINT_E(stderr,
            "ExtSetHostMsg() failed.\n");
        goto EXIT_POINT;
    }

    commInitialized = TRUE;

    /* Free shared memory */
    mbuff_detach("tmpBuf", tmpBuf);
    return SUCCESS;
  
EXIT_POINT:
    /* Free shared memory */
    mbuff_detach("tmpBuf", tmpBuf);
    return(ERROR);

} /* end ProcessConnectMsg */


/* Function: SendMsgHdrToHost ==================================================
 * Abstract:
 *  Send a message header to the host.
 */
PRIVATE boolean_T SendMsgHdrToHost(
    const ExtModeAction action,
    const int           size)  /* # of bytes to follow msg header */
{
    int_T     nSet;
    MsgHeader msgHdr;
    boolean_T error = EXT_NO_ERROR;

    msgHdr.type = (uint32_T)action;
    msgHdr.size = size;

    error = ExtSetHostMsg(extUD,sizeof(msgHdr),(char_T *)&msgHdr,&nSet);
    if (error || (nSet != sizeof(msgHdr))) {
        error = EXT_ERROR;
        PRINT_E(stderr,"ExtSetHostMsg() failed.\n");
        goto EXIT_POINT;
    }

EXIT_POINT:
    return(error);
} /* end SendMsgHdrToHost */


/* Function: SendMsgDataToHost =================================================
 * Abstract:
 *  Send message data to host. You are responsible for sending a header
 *  prior to sending the header.
 */
PRIVATE boolean_T SendMsgDataToHost(const char *data, const int size)
{
    int_T     nSet;
    boolean_T error = EXT_NO_ERROR;

    error = ExtSetHostMsg(extUD,size,data,&nSet);
    if (error || (nSet != size)) {
        error = EXT_ERROR;
        PRINT_E(stderr,"ExtSetHostMsg() failed.\n");
        goto EXIT_POINT;
    }

EXIT_POINT:
    return(error);
} /* end SendMsgDataToHost */


/* Function: SendMsgToHost =====================================================
 * Abstract:
 *  Send a message to the host on the message socket.  Messages can be of
 *  two forms:
 *      o message header only
 *          the type is used as a flag to notify Simulink of an event
 *          that has taken place on the target (event == action == type)
 *      o msg header, followed by data
 */
PRIVATE boolean_T SendMsgToHost(
    const ExtModeAction action,
    const int           size,  /* # of bytes to follow msg header */
    const char          *data)
{
    boolean_T error = EXT_NO_ERROR;

    error = SendMsgHdrToHost(action,size);
    if (error != EXT_NO_ERROR) goto EXIT_POINT;

    if (data != NULL) {
        error = SendMsgDataToHost(data, size);
        if (error != EXT_NO_ERROR) goto EXIT_POINT;
    } else {
        assert(size == 0);
    }

EXIT_POINT:
    return(error);
} /* end SendMsgToHost */


/* Function: ProcessSetParamMsg ================================================
 * Receive and process the EXT_SETPARAM message.
 */
PRIVATE boolean_T ProcessSetParamMsg(const int  msgSize)
{
    const char  *msg;
    boolean_T   error = EXT_NO_ERROR;
    cmdHeader cmdHdr;    

    /* Free shared memory if it has previously been allocated */
    if (msg_shm != NULL ) 
        msgFree();

    /* Allocate shared memory. */
    msg_shm = (char *) mbuff_attach("msg_shm", msgSize); 
    if (msg_shm == NULL) {
        PRINT_E(stderr, "\n!!! Shared memory could not be allocated for 'msg_shm' !!!\n");
        goto EXIT_POINT;
    }

    /*
     * Receive message and set parameters.
     */
    msg = GetMsg(msgSize);
    if (msg == NULL) {
        error = EXT_ERROR; goto EXIT_POINT;
    }

    /* Copy data into shared memory */
    (void) memcpy(msg_shm, msg, msgSize);

    /*
     * Send response to host. 
     */
    error = SendMsgToHost(EXT_SETPARAM_RESPONSE, 0, NULL);
    if (error != EXT_NO_ERROR) goto EXIT_POINT;

    /* Send command to rtl to run SetParam */
    cmdHdr.type = SET_PARAM;  /*actual command to be sent to RTL*/
    cmdHdr.size = msgSize;
    if(write(fd1, &cmdHdr, sizeof(cmdHdr)) < 0) {
        PRINT_E(stderr, "\nThe command SET_PPARAM could not be passed");
        goto EXIT_POINT;
    }

EXIT_POINT:
    return(error);
} /* end ProcessSetParamMsg */

#ifdef NEW
/* Function: ProcessGetParamsMsg ===============================================
 *  Respond to the hosts request for the parameters by gathering up all the
 *  params and sending them to the host.
 */
PRIVATE boolean_T ProcessGetParamsMsg(RTWExtModeInfo *ei)
{
    int_T                         i;
    int_T                         nBytesTotal;
    boolean_T                     error    = EXT_NO_ERROR;
    const DataTypeTransInfo       *dtInfo  = rteiGetModelMappingInfo(ei);
    const DataTypeTransitionTable *dtTable = dtGetParamDataTypeTrans(dtInfo);

    if (dtTable != NULL) {
        /*
         * We've got some params in the model.  Send their values to the
         * host.
         */
        int_T        nTrans   = dtGetNumTransitions(dtTable);
        const uint_T *dtSizes = dtGetDataTypeSizes(dtInfo);

 #ifdef VERBOSE
        PRINT_M("\nUploading initial parameters....\n");
 #endif

        /*
         * Take pass 1 through the transitions to figure out how many
         * bytes we're going to send.
         */
        nBytesTotal = 0;
        for (i=0; i<nTrans; i++) {
            boolean_T tranIsComplex = dtTransGetComplexFlag(dtTable, i);
            int_T     dt            = dtTransGetDataType(dtTable, i);
            int_T     dtSize        = dtSizes[dt];
            int_T     elSize        = dtSize * (tranIsComplex ? 2 : 1);
            int_T     nEls          = dtTransNEls(dtTable, i);
            int_T     nBytes        = elSize * nEls;

            nBytesTotal += nBytes;
        }

        /*
         * Send the message header.
         */
        error = SendMsgHdrToHost(EXT_GETPARAMS_RESPONSE,nBytesTotal);
        if (error != EXT_NO_ERROR) goto EXIT_POINT;

        /*
         * Take pass 2 through the transitions and send the parameters.
         */
        for (i=0; i<nTrans; i++) {
            char_T    *tranAddress  = dtTransGetAddress(dtTable, i);
            boolean_T tranIsComplex = dtTransGetComplexFlag(dtTable, i);
            int_T     dt            = dtTransGetDataType(dtTable, i);
            int_T     dtSize        = dtSizes[dt];
            int_T     elSize        = dtSize * (tranIsComplex ? 2 : 1);
            int_T     nEls          = dtTransNEls(dtTable, i);
            int_T     nBytes        = elSize * nEls;

            error = SendMsgDataToHost(tranAddress, nBytes);
            if (error != EXT_NO_ERROR) goto EXIT_POINT;
        }
    } else {
        /*
         * We've got no params in the model.
         */
        error = SendMsgHdrToHost(EXT_GETPARAMS_RESPONSE,0);
        if (error != EXT_NO_ERROR) goto EXIT_POINT;
    }

EXIT_POINT:
    return(error);
} /* end ProcessGetParamsMsg */
#endif NEW

/*********************
 * Visible Functions *
 *********************/


/* Function: ExtParseArgsAndInitUD =============================================
 * Abstract:
 *  Pass remaining arguments (main program should have NULL'ed out any args
 *  that it processed) to external mode.
 *  
 *  The actual, transport-specific parsing routine (implemented in
 *  ext_svr_transport.c) MUST NULL out all entries of argv that it processes.
 *  The main program depends on this in order to determine if any unhandled
 *  command line options were specified (i.e., if the main program detects
 *  any non-null fields after the parse, it throws an error).
 *
 *  Returns an error string on failure, NULL on success.
 *
 * NOTES:
 *  The external mode UserData is created here so that the specified command-
 *  line options can be stored.
 */
PUBLIC const char_T *ExtParseArgsAndInitUD()
{
    const char_T *error = NULL;

    /*
     * Create the user data.
     */
    extUD = ExtUserDataCreate();

    if (extUD == NULL) {
        error = "Could not create external mode user data.  Out of memory.\n";
        goto EXIT_POINT;
    }

    /*
     * Parse the transport-specific args.
     */
	ExtProcessArgs(extUD);
	if (error != NULL) 
		goto EXIT_POINT;
  
    if (error != NULL) goto EXIT_POINT;

EXIT_POINT:
    if (error != NULL) {
        ExtUserDataDestroy(extUD);
        extUD = NULL;
    }

    return(error);
} /* end ExtParseArgsAndInitUD */


/* Function: ExtWaitForStartMsg ================================================
 * Abstract:
 *  Return true if waiting for host to tell us when to start.
 */
PUBLIC boolean_T ExtWaitForStartMsg(void)
{
    return(ExtWaitForStartMsgFromHost(extUD));
} /* end ExtWaitForStartMsg */


/* Function: rt_UploadServerWork ===============================================
 * Abstract:
 *  Upload model signals to host.
 */
PUBLIC boolean_T rt_UploadServerWork(void)
{
    boolean_T    error = EXT_NO_ERROR;
    int_T        nBytesSent, nBytesPacket, nSet; 
    char_T       *bufPtr;
    char         dummy[4] = {0,0,0,0};

    if ((SendData->status != DISCONNECT) && (SendData->head != SendData->tail)) {
        bufPtr = SendData->net_buff + SendData->tail;

        /* Check whether tail has reached the end of the buffer, marked '0000' */ 
        if (!memcmp((void*)bufPtr, (void *)dummy, 4)) {
            SendData->tail = 0; 
            bufPtr = SendData->net_buff;
        }
        nBytesPacket = *(int *)bufPtr;

        error = ExtSetHostData(extUD, nBytesPacket, bufPtr, &nSet);
        if (error || (nSet != nBytesPacket)) {
            PRINT_E(stderr, "\nExtSetHostData() failed on data upload. \
            \n\tError: %s\n\n",strerror(errno));
        } else {
            SendData->tail += nBytesPacket;
#ifdef RT_MONITORING
            if (SendData->hostNetRate == 1) {
                boolean_T    error = EXT_NO_ERROR;
                error = SendMsgToHost(EXT_HOST_RATE, 0, NULL);
                if (error != EXT_NO_ERROR) 
                    PRINT_E(stderr, "send() failed for EXT_HOST_RATE.\n");

                /* Avoid reentering the this 'if' before the timer expires again. */
                SendData->hostNetRate = 0;  
            }
#endif
        }
    }
    /* If the trigger has been cancelled and there is no more data in the buffer free memory */
    else if ((SendData->status == CANCEL) && (SendData->head == SendData->tail)) {
        cmdHeader cmdHdr;

        SendData->status = DISCONNECT;
        cmdHdr.type = DISCONNECT;
        cmdHdr.size = 0;
        if (write(fd1, &cmdHdr, sizeof(cmdHdr)) < 0) 
            PRINT_E(stderr, "\nThe command DISCONNECT could not be passed. Memory may not be freed");
        }
  
        return SUCCESS;

} /* end rt_UploadServerWork */


/* Function: rt_ExtModeInit ====================================================
 * Abstract:
 *  Called once at program startup to do any initialization related to external
 *  mode. 
 */
PUBLIC boolean_T rt_ExtModeInit(void)
{
    boolean_T error = EXT_NO_ERROR;

    error = ExtInit(extUD);
    if (error != EXT_NO_ERROR) goto EXIT_POINT;

    rtExtModeTestingInit();

EXIT_POINT:
    return(error);
} /* end rt_ExtModeInit */


/* Function: rt_ExtModeSleep ===================================================
 * Abstract:
 *  Called by grt_main, ert_main, and grt_malloc_main  to "pause".  It attempts
 *  to do this in a way that does not hog the processor.
 */
#ifndef VXWORKS
PUBLIC void rt_ExtModeSleep(
    long sec,  /* number of seconds to wait       */
    long usec) /* number of micro seconds to wait */
{
    ExtModeSleep(extUD,sec,usec);
} /* end rt_ExtModeSleep */
#endif

/* Function: rt_MsgServerWork ==================================================
 * Abstract:
 *  If not connected, establish communication of the message line and the
 *  data upload line.  If connected, send/receive messages and parameters
 *  on the message line.
 */
PUBLIC boolean_T rt_MsgServerWork()
{
    MsgHeader  msgHdr;
    boolean_T  hdrAvail;
    boolean_T  error             = EXT_NO_ERROR;
    boolean_T  disconnectOnError = FALSE;
    cmdHeader cmdHdr;

    /*
     * If not connected, attempt to make connection to host.
     */
    if (!connected) {
        rtExtModeTestingKillIfOrphaned(FALSE);

        error = ExtOpenConnection(extUD,&connected);
        if (error != EXT_NO_ERROR) goto EXIT_POINT;
    }

    /*
     * If ExtOpenConnection is not blocking and there are no pending
     * requests to open a connection, we'll still be unconnected.
     */
    if (!connected) goto EXIT_POINT; /* nothing do do */
    
    /*
     * Process messages.
     */

    /* Wait for a message. */
    error = GetMsgHdr(&msgHdr, &hdrAvail);
    if (error != EXT_NO_ERROR) {
        PRINT_W("\nError occured getting message header.");
        disconnectOnError = TRUE;
        goto EXIT_POINT;
    }
    rtExtModeTestingKillIfOrphaned(hdrAvail);
    
    if (!hdrAvail) goto EXIT_POINT; /* nothing to do */

    /*
     * This is the first message.  Should contain the string:
     * 'ext-mode'.  Its contents are not important to us.
     * It is used as a flag to start the handshaking process.
     */
    if (!commInitialized) {
        msgHdr.type = EXT_CONNECT;
    }

    /* 
     * At this point we know that we have a message: process it.
     */
    switch(msgHdr.type) {
    case EXT_HOST_RATE:
    {
        SendData->hostNetRate = msgHdr.size;
        break;
    }
    case EXT_GET_TIME:
    {
		PRINT_VERBOSE(("\nEXT_GET_TIME option has not been implemented in this version of ST-RTL.\n"));
#ifdef MATLAB_V65

        /* Skip verbosity print out - we get too many of these */
        /*PRINT_VERBOSE(("got EXT_GET_TIME message.\n"));*/
        time_T t = rteiGetT(ei);
        
        error = SendMsgToHost(
            EXT_GET_TIME_RESPONSE,sizeof(time_T),
            (char_T *)&t);
        if (error != EXT_NO_ERROR) goto EXIT_POINT;
#endif	  
        break;
    }

    case EXT_ARM_TRIGGER:
    {
        PRINT_VERBOSE(("got EXT_ARM_TRIGGER message.\n"));
        /* Free previous "msg_shm" shared memory allocation. This shared memory is no longer used */
        if (msg_shm != NULL)
        msgFree();
        cmdHdr.type = UPLOAD_ARM_TRIGGER; 
        cmdHdr.size = 0;
        if ( write(fd1, &cmdHdr, sizeof(cmdHdr))<0 ) {
            PRINT_E(stderr,"\nThe command UPLOAD_ARM_TRIGGER could not be passed.\n");
            error = EXT_ERROR;
            goto EXIT_POINT;
        }
#if (WAIT_START_MODEL == 1)
        startModel |= 4; /* Set to one bit 2 */
#endif
        break;
    }

    case EXT_SELECT_SIGNALS:
    {
        const char *msg;

        PRINT_VERBOSE(("got EXT_SELECT_SIGNALS message.\n"));

        msg = GetMsg(msgHdr.size);
        if (msg == NULL) {
            error = EXT_ERROR;
            goto EXIT_POINT;
        }

       /*  error = UploadLogInfoInit(S, msg); */

       /* Free previous "msg_shm" shared memory allocation */
       msgFree();

       /* Allocate shared memory */
       msg_shm = (char *)mbuff_attach("msg_shm", msgHdr.size); 
       if (msg_shm == NULL) {
           PRINT_E(stderr,"\n!!! 'msg_shm' could not be allocated in shared memory!!!\n");
           error = EXT_ERROR; 
           goto EXIT_POINT;
       }

       /* Copy data to shared memory that 'updown' can read it */
       (void)memcpy(msg_shm, msg, msgHdr.size);

       /* Send command to rtl to run SetParam */
       cmdHdr.type = UPLOAD_LOG_INFO_INIT;  /*actual command to be sent to RTL*/
       cmdHdr.size = msgHdr.size;
       if(write(fd1, &cmdHdr, sizeof(cmdHdr)) < 0) {
          PRINT_E(stderr, "\nThe command UPLOAD_LOG_INFO_INIT could not passed.");
          error = EXT_ERROR; 
          goto EXIT_POINT;   
        }
        break;
    }

    case EXT_SELECT_TRIGGER: 
    {
        const char *msg;

        PRINT_VERBOSE(("got EXT_SELECT_TRIGGER message.\n"));

        msg = GetMsg(msgHdr.size);
        if (msg == NULL) {
            error = EXT_ERROR;
            goto EXIT_POINT;
        }

        /*  error = UploadInitTrigger(S, msg); */
     
        /* Free previous "msg_shm" shared memory allocation */
        msgFree();
      
        /* Allocate shared memory */
        msg_shm = (char *) mbuff_attach("msg_shm", msgHdr.size); 
        if (msg_shm == NULL) {
            PRINT_E(stderr, "\n!!! 'msg_shm' could not be allocated in shared memory!!!\n");
            error = ERROR; 
            goto EXIT_POINT;
        }

        /* Copy data to shared memory that updown can read it */
        (void)memcpy(msg_shm, msg, msgHdr.size);

        /* Send command to rtl to run SetParam */
        cmdHdr.type = UPLOAD_INIT_TRIGGER;  /*actual command to be sent to RTL*/
        cmdHdr.size = msgHdr.size;
        if(write(fd1, &cmdHdr.type, sizeof(cmdHdr)) < 0) {
            PRINT_E(stderr, "\nThe command UPLOAD_INIT_TRIGGER could not be passed.\n");
            error = ERROR; 
            goto EXIT_POINT;   
        }
        if (error != EXT_NO_ERROR) {
            PRINT_E(stderr,"\nError in UploadInitTrigger\n");
            goto EXIT_POINT;
        }
        break;
    }

    case EXT_CONNECT:
    {
        PRINT_VERBOSE(("got EXT_CONNECT message.\n"));
        /* error = ProcessConnectMsg(S); */
        cmdHdr.type = PROCESS_CONNECT_MSG_RTL;
        cmdHdr.size = 0;
        if(write(fd1,&cmdHdr, sizeof(cmdHdr)) < 0) {
            PRINT_E(stderr, "\nThe command PROCESS_CONNECT_MSG_RTL could not be passed.\n");
            error = ERROR;
            goto EXIT_POINT;
        }

#if (WAIT_START_MODEL == 1)
        startModel |= 1; /* Set to one bit 0. Keep value of bit 4 in case the RT thread was started. */
#endif
        break;
    }

    case EXT_SETPARAM:
    {
        PRINT_VERBOSE(("got EXT_SETPARAM message.\n"));
        error = ProcessSetParamMsg(msgHdr.size);
        if (error != EXT_NO_ERROR) goto EXIT_POINT;
        break;
    }

    case EXT_GETPARAMS:
    {
		PRINT_VERBOSE(("\nEXT_GETPARAMS option has not been implemented in this version of ST-RTL.\n"));
#ifdef MATLAB_V65
        PRINT_VERBOSE(("got EXT_GETPARAMS message.\n"));
        error = ProcessSetParamMsg(msgHdr.size);
        if (error != NO_ERR) {
             PRINT_E(stderr, "\n!!! An error occurred in ProcessSetParamMsg !!!\n");
             error = ERROR;
             goto EXIT_POINT;
        }
#endif /* MATLAB_V65*/	  
        break;
    }

    case EXT_DISCONNECT_REQUEST:
    {
        PRINT_VERBOSE(("got EXT_DISCONNECT_REQUEST message.\n"));
        
        /*
         * Note that from the target's point of view this is
         * more a "notify" than a "request".  The host needs to
         * have this acknowledged before it can begin closing
         * the connection.
         */
        error = SendMsgToHost(EXT_DISCONNECT_REQUEST_RESPONSE, 0, NULL);
        if (error != EXT_NO_ERROR) goto EXIT_POINT;

        /* Send command to rtl to disconnect */
        cmdHdr.type = DISCONNECT;
        cmdHdr.size = 0;
        if (write(fd1, &cmdHdr, sizeof(cmdHdr)) < 0) {
            PRINT_E(stderr, "\nThe command DISCONNECT could not be passed.");
            error = ERROR; 
            goto EXIT_POINT;  
        }
#if (WAIT_START_MODEL == 1)
        startModel &= 14; /* Set to zero bit 0, keep value of bit 4 in case the RT thread was started */
#endif

        SendTerminatingMessage();
        DisconnectFromHost();
        SendData->status = DISCONNECT;
        break;
    }
#if (WAIT_START_MODEL == 1)
    case EXT_MODEL_START:
        PRINT_VERBOSE(("got EXT_MODEL_START message.\n"));

        modelStatus = TARGET_STATUS_RUNNING;
        startModel |= 2; /* Set to one bit 1 */ 

        error = SendMsgToHost(EXT_MODEL_START_RESPONSE, 0, NULL);
        if (error != EXT_NO_ERROR) goto EXIT_POINT;
        break;
#endif /*  WAIT_START_MODEL */

    case EXT_MODEL_STOP:
        PRINT_VERBOSE(("got EXT_MODEL_STOP message.\n"));
        cmdHdr.type = STOP_MODEL;
        cmdHdr.size = 0;
        if (write(fd1, &cmdHdr, sizeof(cmdHdr)) < 0) {
            PRINT_E(stderr, "The command STOP_MODEL could not be passed\n");
            error = ERROR;
            goto EXIT_POINT;
        }
        break;

    case EXT_MODEL_PAUSE:
	  PRINT_VERBOSE(("\nEXT_MODEL_PAUSE option has not been implemented in this version of ST-RTL.\n"));
#ifdef MATLAB_V65    
        PRINT_VERBOSE(("got EXT_MODEL_PAUSE message.\n"));
        modelStatus = TARGET_STATUS_PAUSED;
        startModel  = FALSE;

        error = SendMsgToHost(EXT_MODEL_PAUSE_RESPONSE, 0, NULL);
        if (error != EXT_NO_ERROR) goto EXIT_POINT;
#endif /* MATLAB_V65 */
        break;

    case EXT_MODEL_STEP:
        PRINT_VERBOSE(("\nEXT_MODEL_STEP option has not been implemented in this version of ST-RTL.\n"));
#ifdef MATLAB_V65
        PRINT_VERBOSE(("got EXT_MODEL_STEP message.\n"));
        if ((modelStatus == TARGET_STATUS_PAUSED) && !startModel) {
            startModel = TRUE;
        }
#endif     
        error = SendMsgToHost(EXT_MODEL_STEP_RESPONSE, 0, NULL);
        if (error != EXT_NO_ERROR) goto EXIT_POINT;

        break;

    case EXT_MODEL_CONTINUE:
        PRINT_VERBOSE(("\nEXT_MODEL_CONTINUE option has not been implemented in this version of ST-RTL.\n"));
  
#ifdef MATLAB_V65
        PRINT_VERBOSE(("got EXT_MODEL_CONTINUE message.\n"));
        if (modelStatus == TARGET_STATUS_PAUSED) {
            modelStatus = TARGET_STATUS_RUNNING;
            startModel  = FALSE;
        }
#endif /* MATLAB_V65 */       
        error = SendMsgToHost(EXT_MODEL_CONTINUE_RESPONSE, 0, NULL);
        if (error != EXT_NO_ERROR) goto EXIT_POINT;

        break;

    case EXT_CANCEL_LOGGING:
        PRINT_VERBOSE(("got EXT_CANCEL_LOGGING message.\n"));
#if (WAIT_START_MODEL == 1)
        if (startModel >= 8) { /* RT thread is running if bif 4 of startModel is one */
             SendData->status = HOLD;  /* Only HOLD if the RT thread was running */
             startModel &= 11; /* Set to zero bit 2 only if the RT thread was running. Keep value of bit 4. */
       }
#else
       SendData->status = HOLD;
#endif

        /*  UploadCancelLogging();  */
        break;

    default:
        PRINT_E(stderr,"received invalid message.\n");
        break;
    } /* end switch */

EXIT_POINT:
    if (error != EXT_NO_ERROR) {
        if (disconnectOnError) {
            PRINT_E(stderr,
                "Error occured in rt_MsgServerWork.\n"
                "Disconnecting from host!\n");
            DisconnectFromHost();
        }
    }

    return(error);
} /* end rt_MsgServerWork */


/* Function: rt_ExtModeShutdown ================================================
 * Abstract:
 *  Called when target program terminates to enable cleanup of external 
 *  mode.
 */
PUBLIC boolean_T rt_ExtModeShutdown(void)
{
    boolean_T           error              = EXT_NO_ERROR;
    boolean_T           dataPending        = 0; /*IsAnyDataReadyForUpload();*/
    const ExtModeAction shutdownAction     = (dataPending) ?
        EXT_MODEL_SHUTDOWN_DATA_PENDING : EXT_MODEL_SHUTDOWN;
    const char          *shutdownActionStr = (dataPending) ?
        "EXT_MODEL_SHUTDOWN_DATA_PENDING" : "EXT_MODEL_SHUTDOWN";

    /*
     * Make sure buffers are flushed so that the final points get to
     * host (this is important for the case of the target reaching tfinal
     * while data is uploading is in progress).
     */
    /*UploadPrepareForFinalFlush();*/
    rt_UploadServerWork();

    /* UploadLogInfoTerm(); */
    if (msgBuf != NULL) free(msgBuf);
    
    if (connected) {
        error = SendMsgToHost(shutdownAction, 0, NULL);
        if (error != EXT_NO_ERROR) {
            PRINT_E(stderr,
                "\nError sending '%s' message to host.\n", shutdownActionStr);
        }
        connected = FALSE;
        commInitialized = FALSE;
        modelStatus = TARGET_STATUS_WAITING_TO_START;        
    }

    ExtShutDown(extUD);
    ExtUserDataDestroy(extUD);

    rtExtModeTestingRemoveBatMarker();
        
    return(error);
} /* end rt_ExtModeShutdown */

PUBLIC void SendTerminatingMessage()
{
	boolean_T error = EXT_NO_ERROR;
	char_T    buff[8] = {8, 0, 0, 0, 5, 0, 0, 0} ;    
	int_T     nSet;        

  	/* The terminating message consists of 
	 *   8: number of bytes 
	 *   5: value of TRIGGER_TERMINATING
	 */
	
	if (SendData->status != DISCONNECT) {
		error = ExtSetHostData(extUD, 8, buff, &nSet);
		if (error || (nSet != 8)) 
			PRINT_E(stderr, "\nTerminating message could not be sent %d.\n",error);
	}
}
#ifdef eRASE
/* Function ====================================================================
 * Pause the process (w/o hogging the cpu) until receive start message
 * from the host.  The message/upload server must continue to process
 * events (otherwise the host would not be able to communicate with the target).
 */
void rtExtModeWaitForStartMsg(RTWExtModeInfo  *ei,
                              boolean_T       *stopReqPtr)
{
    /*
     * Pause until receive model start message.
     */
    if (ExtWaitForStartMsg()) {
        while(!startModel && !(*stopReqPtr)) {
            rt_ExtModeSleep(0L, 375000L);
            rt_MsgServerWork();
            rt_UploadServerWork();
        }
    }
    if (modelStatus != TARGET_STATUS_PAUSED) {
        modelStatus = TARGET_STATUS_RUNNING;
    } else {
        /* leave in pause mode */
    }
}
#endif

/* [EOF] ext_svr.c */
