/*******************************************************************
 * Simulink Target for Real-Time Linux (ST-RTL)                    *
 *                                                                 *
 * Author:       Raul Murillo Garcia <r.murillogarcia@gcal.ac.uk>  *
 * Last review:  30 August 2002                                    *
 *                                                                 *
 * This file is the execution engine in kernel space.              *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of the GNU General Public License     *
 *                                                                 *
 *******************************************************************/
 
#include <linux/module.h>
	
MODULE_AUTHOR ("Raul Murillo Garcia, <r.murillogarcia@gcal.ac.uk>");
MODULE_DESCRIPTION ("Simulink Target for Real-Time Linux (ST-RTL). \
				This module runs a real-time controller designed \
				with Simulink and converted into C code with \
				Real-Time Workshop.");

/*****************/
/* Header files  */
/*****************/

/* rtl header files */ 
#include <rtl.h> 
#include <rtl_fifo.h>
#include <rtl_time.h>
#include <pthread.h>

/* ST-RTL headers */
#include "common.h"
#include "krnl_main.h"
#include "ciodas08jrao.h"

/* RTW header files */ 
#include "simstruc.h" 
#include "ext_svr_rtl.h"
#include "tmwtypes.h"
#include "updown.h"
#include "rt_logging.h"
#include "rtmodel.h"

#if defined(MULTITASKING)
# define NUM_RT_THREADS (NUMST - TID01EQ) /* Number of Real-Time Threads */
#else
# define NUM_RT_THREADS 1 
#endif

#define printf rtl_printf


/************************/ 
/* External Functions    */
/************************/
extern RT_MODEL *MODEL(void);                     /* Defined in model.reg */
extern int_T model_init(RT_MODEL *S);            /* Defined in grt_main.c */
extern void  model_run(RT_MODEL *S, int_T tid);  /* Defined in grt_main.c */
extern int_T model_clean(RT_MODEL *S);           /* Defined in grt_main.c */

/*********************/ 
/*  Local Functions  */
/*********************/
void *rt_thread_func(void *arg);
void *da_thread_func(void *arg);
void *watchdog_func(void *arg);

/************************/ 
/* Global variables     */
/************************/
static RT_MODEL *S;

pthread_t rtThread[NUM_RT_THREADS],  /* Real-time threads.           */
          watchdogThread;            /* Watchdog thread to make sure time cycles are met. */	    

/* The rt threads may not be cancelled in stop_model(). If so, this task has to be done in           *
 * cleanup_module(). rtThreadCreated tells us wheather a rt thread has been creatd or cancelled.     */
 static int_T  rtThreadCreated[NUM_RT_THREADS],
              watchdogThreadCreated;

/* Flags used by the rt-tasks and the watchdog to detect overrun occurred. */
static unsigned int overrunTask[NUM_RT_THREADS];

#ifdef RT_MONITORING
pthread_t     daThread;         /* Discarding algorithm thread. */	    
static int_T  daThreadCreated;  /* daThreadCreated tell us if the thread to 
                                   run the Discarding Algorithm has been created. */
#endif


/*********************/
/*  LOCAL FUNCTIONS  */
/*********************/

/* Function: create_rt_threads() ===========================================================
 * Creates the rt threads and the DA thread.
 */
static int_T create_rt_threads(void)
{
	int_T tid, retval = SUCCESS;

	memset(overrunTask, 0, sizeof(int) * NUM_RT_THREADS);

	/* Create watchdog thread  */
	retval = pthread_create(&watchdogThread, NULL, watchdog_func, (void*) NULL);
	if (retval != 0)  {
		PRINT_W("\n!!! The Watchdog thread could not be created !!!\n");
		watchdogThreadCreated = FALSE;
		retval = ERROR;
	} else 
		watchdogThreadCreated = TRUE;

	/* Create the Real-Time threads */
	for (tid = 0; tid < NUM_RT_THREADS; tid++) {
		retval = pthread_create(&rtThread[tid], NULL, rt_thread_func, (void*) tid);
		if (retval != 0)  {
			PRINT_W("\n!!! Real-Time thread (tid # %d) could not be created !!!\n", tid);
			rtThreadCreated[tid] = FALSE;
			retval = ERROR;
		} else {
			rtThreadCreated[tid] = TRUE;
			/* Continuous blocks (tid 0) run at the base sample rate time. If one of these 
			 * blocks is present tid 1 defines the base sample rate, tid 1, the next faster 
			 * sample rate and so on. Therefore tid 0 and tid 1 have the same value. One 
			 * needs to be ommit.The preprocessor macro TID01EQ is used to discern this
			 */
		}
	}
	
#ifdef RT_MONITORING
	/* Create the Discarding Algorithm thread */
	retval = pthread_create(&daThread, NULL, da_thread_func, (void*) NULL);
	if (retval != 0)  {
		PRINT_W("\n!!! The Discarding Algorithm thread could not be created !!!\n");
		daThreadCreated = FALSE;
		retval = ERROR;
	} else 
		daThreadCreated = TRUE;
#endif

	/* Make the threads periodic. */
	{
		hrtime_t tstart, bsperiod;

		/* Threads start after a second. */
		tstart = gethrtime() + NSECS_PER_SEC;
		/* The watchdog starts with an offset of 1 base sample period. */ 
		bsperiod = (hrtime_t) (NSECS_PER_SEC*rtmGetSampleTime(S,0 + TID01EQ));

		/* Make the watchdog periodic. */
		pthread_make_periodic_np(watchdogThread, tstart + bsperiod , bsperiod); 
	
		/* Make the threads periodic. */
		for (tid = 0; tid < NUM_RT_THREADS; tid++) {
			pthread_make_periodic_np(rtThread[tid], tstart, \
							(hrtime_t)*(NSECS_PER_SEC * rtmGetSampleTime(S,tid + TID01EQ)));
		}

#ifdef RT_MONITORING
		pthread_make_periodic_np(daThread, tstart, NSECS_PER_SEC/DA_RATE);
#endif			
	}

	return retval;

} /* End of create_rt_threads()  */


/* Function: cancel_rt_threads() ===========================================================
 * Cacels the rt threads and the DA thread..
 */
static void cancel_rt_threads(void)
{
	int_T tid;

	/* Cancel watchdog thread. */
	if (watchdogThreadCreated == TRUE)  {
		pthread_cancel(watchdogThread);
		/* In case cancel_rt_threads() is called again before the thread has *
		 * been completly cancelled. It would wait twice for a single thread. */
		watchdogThreadCreated = CANCEL;  
	}
	/* Cancel rt-threads */
	for (tid = 0; tid < NUM_RT_THREADS; tid++) {
		if (rtThreadCreated[tid] == TRUE)  {
			pthread_cancel(rtThread[tid]);
			rtThreadCreated[tid] = CANCEL;   
		}
	}
	
	/* Wait for the threads to be cancelled. */
	if (watchdogThreadCreated == CANCEL)  {
		/* In case cancel_rt_threads() is called again before the thread has
		 * been completly cancelled. */	
		watchdogThreadCreated = HOLD;
		pthread_join(watchdogThread, NULL);
		watchdogThreadCreated = FALSE;   
	}
	for (tid = 0; tid < NUM_RT_THREADS; tid++) {
		if (rtThreadCreated[tid] == CANCEL)  {
			rtThreadCreated[tid] = HOLD; 
			pthread_join(rtThread[tid], NULL);  /* Wait for rt thread to be cancelled. */
			rtThreadCreated[tid] = FALSE;       /* The rt thread has been cancelled.   */
		}
	}
	
#ifdef RT_MONITORING
	if (daThreadCreated == TRUE)  {
		pthread_cancel(daThread);
		daThreadCreated = CANCEL;  
		pthread_join(daThread, NULL);
		daThreadCreated = FALSE;   
	}
#endif

} /* End of cancel_rt_threads()  */


/***********************/
/*  GLOBAL  FUNCTIONS  */
/***********************/

/* Function: watchdog_func() ==============================================================
 * Watchdog thread to ensure that the task periods are not too small. It is invoked at the  
 * highest sample rate. Thread are checked accoring to their tid hit. It has the highest 
 * priority to be able to cancel the rest of threads if necessary. The process is stopped 
 * if the thread did not reset their corresponding flag.Two overrun condition must occur to
 * to signal this exeption. We do this because the task execution period is always larger 
 * the first time it is executed.
 */
void *watchdog_func(void *arg)
{
	struct sched_param p;
	int_T  *sampleHit = rtmGetSampleHitPtr(S);
	unsigned int overrunWdog[NUM_RT_THREADS];
	/* Overrun condition. */
	char overrun [NUM_RT_THREADS];
	
	memset(overrun, 0, sizeof(char) * NUM_RT_THREADS);
	memset(overrunWdog, 0, sizeof(int) * NUM_RT_THREADS);
	
	/* Enable floating point */
	pthread_setfp_np(pthread_self(), 1); 

	/* Highest priority. */
	p.sched_priority = NUM_RT_THREADS + 2; 
   
	pthread_setschedparam(pthread_self(), SCHED_FIFO, &p);
	
	while(1)
	{
		register int i;
	
		pthread_wait_np();
		for (i = 0; i < NUM_RT_THREADS; i++) {
			/*  Is it a tid hit?  */
			if (sampleHit[i+TID01EQ]) {
		             /* Counter mismatch? */
				if (++overrunWdog[i] > overrunTask[i]) {
					/* Second consecutive overrun? */
					if (overrun[i]++) {
						stop_model();
						PRINT_W("\n!!! The model specifications are too fast for this hardware !!!\n");							
						break;
					}
				} else {
					overrun[i] = 0;
				}
			}
		}
		pthread_testcancel(); /* Finish thread here if cancelled from stop_model() */
	}
	
} /* End of watchdog_func() */


/* Function: rt_thread_func() ==============================================================
 * This is the handler for the rt threads. Gives priority and makes the rt thread periodic.
 * Each tid calls upon the contrller cycle at a specifc rate.
 */
void *rt_thread_func(void *arg)
{  
	int_T tid = (int_T) arg;
	struct sched_param p;

	/* Enable floating point */
	pthread_setfp_np(pthread_self(), 1); 

	/* The smaller the tid the higher the sample rate and therefore  *
	 * the higher the priority. Lowest priority is the discarding algorithm.
	 * Hence the +1. */
	p.sched_priority = NUM_RT_THREADS - tid + 1; 
   
	pthread_setschedparam(pthread_self(), SCHED_FIFO, &p);
  
	while (1)  {
		pthread_wait_np();
		/* Run model */ 
		model_run(S, tid); 
		/* Increase counter */
		overrunTask[tid]++;
	
		pthread_testcancel(); /* Finish thread here if cancelled from stop_model() */
		/* By default a thread is PTHREAD_CANCEL_ENABE and PTHREAD_CANCEL_DEFERRED */
	}

}  /* End of rt_thread_func() */

#ifdef RT_MONITORING
/* Function: da_thread_func() ==============================================================
 * This is the handler for the function that runs the Discarding Algorithm and updates 
 * the Discarding Factor. The periodicity of this thread is defined by the macro 
 * DA_RATE. The default value is 100, i.e. 10 ms.
 */
void *da_thread_func(void *arg)
{
	struct sched_param p;
 
	pthread_setfp_np(pthread_self(), 1); /* Allow for floating point */

	/* Lowest priority */
	p.sched_priority = 1; 
   
	pthread_setschedparam(pthread_self(), SCHED_FIFO, &p);

	while (1)  {
		pthread_wait_np();
		df_update(); /* Run the Discarding Algorithm and update the Discarding Factor.*/
		pthread_testcancel(); /* Finish thread here if cancelled from stop_model() */
		/* By default a thread is PTHREAD_CANCEL_ENABE and PTHREAD_CANCEL_DEFERRED */
	}
}  /* End of da_thread_func() */
#endif

/* Function: stop_model ===========================================================
 * Stops the real-time process by cancelling the real-time threads. The model is 
 * cleaned and shared memory freed. Also used in grt_main.c.
 */
void  stop_model(void)
{
	static int stopping = FALSE; 
	
	if (!stopping) {
		int_T retval;
		cmdHeader cmdHdr;
    
		/* All the rt threads are killed */
		cancel_rt_threads();

		/* Clean the model */
		retval = model_clean(S);

		if (retval == ERROR) 
			PRINT_W("\n!!!! Error on finishing the model !!!!\n");
	  
#ifdef EXT_MODE
		tmpBufFree(); /* In case an error occurred durnih handshaking */
#endif
	
		PRINT_M("\n*** Model finished\n");

		cmdHdr.type =  STOP_MODEL;    
		cmdHdr.size = 0;
		rtf_put(RTF0, &cmdHdr, sizeof(cmdHdr));
	}
	
} /* End of stop_model()  */


/* Function: fifo_handler ===================================================================
 * This function is invoked when data is read or write on FIFO #1 by the user space process.
 * Commands are retrieved and an action is taken accordingly.
 */
int fifo_handler(unsigned int fifo) 
{
	int_T retval;
	cmdHeader cmdHdr;	
    
	/* At this point there should be data on the FIFO written by us_main */
	rtf_get(RTF1, &cmdHdr, sizeof(cmdHdr)); 
    
	switch (cmdHdr.type){ 
		case STOP_MODEL:
			stop_model(); 
			break;
    
#ifdef EXT_MODE
		case DISCONNECT:
			EndLoggingSession();
			break;

		case PROCESS_CONNECT_MSG_RTL:
			/* if the value return by ProcessConnectMsg_rtl(S) is not ERROR, then is the     *
			* number of bytes allocated in shared memory. This number has to be passed up   *
			* to the 'us application'                                                       */
			retval = ProcessConnectMsg_rtl(rtmGetRTWExtModeInfo(S));
			if (retval == ERROR) {
				PRINT_W("\n!!!! Error in function ProcessConnectMsg_rtl() !!!!\n");
				goto SEND_ERROR;
			}
			cmdHdr.type = PROCESS_CONNECT_MSG;  
			cmdHdr.size = retval;
			retval      = rtf_put(RTF0, &cmdHdr, sizeof(cmdHdr));
			if (retval < 0) {
				PRINT_W("\n!!!! Could not write data from FIFO \n");
				goto SEND_ERROR;
			}
			break;
    
		case SET_PARAM:  
			/* Shared memory used in the ProcessConnectMsg must be freed. That function  *
			 * is run right before this point                                            */
			tmpBufFree();
			SetParam(rtmGetRTWExtModeInfo(S), cmdHdr.size); /*send the size of message header*/
			break;
   
		case UPLOAD_LOG_INFO_INIT:
			retval = UploadLogInfoInit(S, cmdHdr.size);
			if (retval == ERROR) {
				PRINT_W("\n!!!! Error in function UploadLogInfoInit() !!!!\n");
				goto SEND_ERROR;
			}
			break;
    
		case UPLOAD_INIT_TRIGGER:
			retval = UploadInitTrigger((RTWExtModeInfo *)rtmGetRTWExtModeInfo(S), cmdHdr.size);
			if (retval == ERROR) {
				PRINT_W("\n!!!! Error in function UploadInitTrigger() !!!!\n");
				goto SEND_ERROR;
			}
			break;
    
		case UPLOAD_ARM_TRIGGER:
			/* Free the previous allocation. */
			SendDataBuffFree();
			/* Allocate buffer (SendData) in shared memory */
			if (SendDataBuffAlloc() == ERROR) {
				PRINT_W("\n!!!! Shared Memory could not be allocated for buffer !!!!\n");
				goto SEND_ERROR;
			}
			cmdHdr.type = SEND_DATA;  
			cmdHdr.size = 0;
			retval     = rtf_put(RTF0, &cmdHdr, sizeof(cmdHdr));
			if (retval < 0) {
				PRINT_W("\n!!!! Could not write data from FIFO \n");
				goto SEND_ERROR;	
			}
			
			SendDataInit();		
			UploadArmTrigger(); 
			PRINT_M("\n*** Handshaking finished.\n");				
			break;
    
#endif /* EXT_MODE */
    
		case START_MODEL:
			retval = create_rt_threads();
			if (retval == ERROR)
				goto SEND_ERROR;
			break;

		default:
			PRINT_W("\nCommand read from FIFO was not recognised"
					"\n\tCommand sent: %d\n", cmdHdr.type);
			return -1;
			break;
	} /* Switch */

	return 0;
  
SEND_ERROR:
	stop_model();
	return 0;
} /* End of fifo_handler() */

/* Function: init_module ===========================================================
 * init_module() is invoked when the module 'model_rtl.o' is inserted in the kernel.
 * FIFOs are created, shared memory allocated and model initialized.
 */
int init_module(void)
{
	int_T retval = SUCCESS;
	cmdHeader cmdHdr;
	
	cmdHdr.type = SUCCESS;
	cmdHdr.size = 0;
    
     /* Don't export any global variables. From kernel version 2.5.0 *
 	* onwards symbols are not exported by default.                 */
#if LINUX_KERNEL_CODE < KERNEL_VERSION(2,5,0)
	EXPORT_NO_SYMBOLS;
#endif
	rtf_destroy(RTF0);  
	rtf_destroy(RTF1);  
    
	/* If an error occurs here, the error cannot be sent to us_main through FIFO, hence return -1 */
	if ((retval = rtf_create(RTF0, RTF0_SIZE)) != 0)  {
		PRINT_W("\n!!!! FIFO rtf%d could not be created.!!!!\n\tError: %d\n", RTF0, retval);
		return -1;
	}
	if ((retval = rtf_create(RTF1, RTF1_SIZE)) != 0)  {
		PRINT_W("\n!!!! FIFO rtf%d could not be created: !!!!\n\tError: %d\n", RTF1, retval);
		cmdHdr.type = ERROR; goto EXIT_POINT;
	}
  
	/* fifo_handler will run anytime data is read from or written to FIFO RTF1. 
	* This FIFO  is used to send data from us_main to krnl_main */
	if ((retval = rtf_create_handler(RTF1, &fifo_handler)) != 0)  {
		PRINT_W("\n!!!! fifo_handler could not be created. !!!!\n\tError: %d\n", retval);
		cmdHdr.type = ERROR; goto EXIT_POINT;
	}
  
#ifdef EXT_MODE
	/* Initialize shared memory */
	retval = SendDataAlloc();
	if (retval == ERROR) {
		PRINT_W("\n!!!! Shared Memory could not be allocated !!!!\n");
		cmdHdr.type = ERROR; goto EXIT_POINT;
	}
#endif 
  
	PRINT_M("\n*** Initializing the model... \n");

	/* Get address of SimStruct S defined in model.reg / model() */
	S = MODEL();

	/* Initialize model */ 
	retval = model_init(S);
	if (retval == ERROR) { 
		PRINT_W("\n!!!! Model could not be initialized !!!!\n");
		model_clean(S);
		cmdHdr.type = ERROR; goto EXIT_POINT; 
	}
	
	/* Detect Data Acquisition Card. */
	if (das08jrao_init() == -ENODEV)
		PRINT_W("\nData Acqusition Card CIO-DAS08/JR-AO not detected \n");
	
EXIT_POINT:
	
	retval = rtf_put(RTF0, &cmdHdr, sizeof(cmdHdr));
	if (retval < 0) {
		PRINT_W("\n!!!! Could not wrtie data from FIFO \n");
		return -1;
	}	
	
	return 0;

} /* End of init_module() */

/* Function: cleanup_module ===========================================================
 * cleanup_module() is invoked when the module 'model_rtl.o' is unloaded from the 
 * kernel. FIFOs are destroyed, the real-time cancelled and the shared memory freed.
 */
void cleanup_module(void) 
{
	/* All the rt threads are killed if this task has not been carried out in stop_model. */
	cancel_rt_threads();

	rtf_destroy(RTF0); 
	rtf_destroy(RTF1);  

#ifdef EXT_MODE
	SendDataFree();
#endif
	
	/* Free shared memory for data logging. */
	freeLogSM ();

	/* Relsease driver */
	das08jrao_close();
	
} /* End of cleanup_module() */
