/*******************************************************************
 * Simulink Target for Real-Time Linux (ST-RTL)                    *
 *                                                                 *
 * Author:       Raul Murillo Garcia <r.murillogarcia@gcal.ac.uk>  *
 * Last review:  30 August 2002                                    *
 *                                                                 *
 * This file is the execution engine in user space.                *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of the GNU General Public License     *
 *                                                                 *
 *******************************************************************/
 
#include <sys/time.h> 
#include <signal.h>
#include <stdio.h>
#include <fcntl.h>

/*RTW header files  */
#include "tmwtypes.h"
#include "simstruc.h" 
#include "ext_svr_us.h"

/* ST-RTL header */
#include "common.h"  

#ifndef SAVEFILE
# define MATFILE2(file) #file ".mat"
# define MATFILE1(file) MATFILE2(file)
# define MATFILE MATFILE1(MODEL)
#else
# define MATFILE QUOTE(SAVEFILE)
#endif

#define RUN_FOREVER -1.0

#if (WAIT_START_MODEL == 1)
/* Wait for START message from Simulink side.Three conditions are needed to create the
 * RT thread. Host-Target need to be connected, trigger armed and model started. When one
 * of this conditions is true we set to high bit 0, 1 or 2 of startModel. Therefore, when the 
 * value is ...0000111 the thread can be staretd. On the other hand when a disconnection occurs
 * or the trigger is cancelled we set to low the corresponding bit.
 * CONNECT: bit 0
 * START:   bit 1
 * ARM:     bit 2
 */
int_T startModel = 0;
#endif

volatile char *dlogbuff = NULL;
int_T fd0,fd1; 

/* Function: exit_func =================================================================
 * Terminates the user spalce process. Network sockets are closed, data logging files 
 * are closed, memory is freed and FIFOs are released. Finally, the last 8 messages 
 * from the kernel are printed. This is handy in case an error has occurred.
 */
static void exit_func(void)
{
	int_T retval;

#ifdef EXT_MODE
	/* Send the terminating message which is at the end of the buffer */
	SendTerminatingMessage();

	/* Close sockets */
	retval = rt_ExtModeShutdown(); 
	if (retval == ERROR) {
		PRINT_E(stderr,"\n!!! Error in function rt_ExtModeShutdown. !!!\n");
	}

	msgFree();  /* Free shared memory in case an error occurred in krnl_main */
	msgFree();  /* Free shared memory in case an error occurred in krnl_main */
	SendDataFree(); /* Free shared memory for strucuture SendData  */

#endif

	close(fd0); 
	close(fd1);

	/* Unload modules from the kernel */
#define P(x) #x
#define SYSTEM(x) system("make -f " P(x) ".mk unload")
	SYSTEM(MODEL);
	
#define P(x) #x
#define PRINTF(x) PRINT_M("\n ### Model "P(x)" stopped \n")
	PRINTF(MODEL);

	/* Print the last 8 messages from the kernel. */
	PRINT_M("\nMessages from the kernel...\n\n");
	system("dmesg | tail -8");

} /* End of exit_func() */


/* Function: stop_model =================================================================
 * Sends command STOpMODEL to the the real-time process to terminate the application. 
 */
static stop_model(void)
{
	cmdHeader cmdHdr;
	
	cmdHdr.type = STOP_MODEL;
	cmdHdr.size = 0;
	if (write(fd1, &cmdHdr, sizeof(cmdHdr)) < 0) {
		PRINT_E(stderr, "\nThe command STOP_MODEL could not be passed. ");
		exit(-1);
	}
}  /*  End of stop_model()  */


/* Function: doDataLogging=====================================================
 * Abstract: 
 *   Gets data logging from shared memory and creates a MAT-file
 */
static int doDataLogging(int nbytes)
{
	FILE    *fptr;
	int_T   retval  = SUCCESS,
	        filesize;
	char    *buf;

	/* Allocate shared memory. */
	dlogbuff = (volatile char *) mbuff_alloc("dlogbuff", nbytes);
	if (dlogbuff == NULL) {
		PRINT_E(stderr,"\nError: Shared memory for data logging could not be allocated");
		retval = ERROR;	goto EXIT_POINT;
	}	
	
	memcpy(&filesize, dlogbuff, sizeof(int_T));
	buf = dlogbuff + sizeof(int_T);

	/*******************************
	* Create MAT file with header *
	*******************************/

	if ((fptr=fopen(MATFILE,"wb")) == NULL) {
		PRINT_E(stderr,"*** Error opening %s", MATFILE);
		retval = ERROR; goto EXIT_POINT;
	}
	
	/* Put data on shared memory. */
	if (fwrite(buf, 1, filesize, fptr) != filesize) {
		PRINT_E(stderr,"\n!!! Error when writing MAT-file");
		retval = ERROR;
	}	
	
	fclose(fptr);		

 EXIT_POINT:	
	/* Free shared memory.*/ 
	if (dlogbuff != NULL)       
		mbuff_free("dlogbuff", dlogbuff);

	return retval;

 } /* end doDataLogging */
 

/* Function: main =================================================================
 * Main function.User-defined parameters are checked, FIFOs created, shared memory
 * allocated, network sockets opened, data logging files opened, etc. Once the 
 * initialization is finished, the process enters an endless loop where commands 
 * from both the real-time process (FIFOs) and host machine (network) are read.
 */
int main (void) 
{
	boolean_T  retval;
	cmdHeader  cmdHdr;
	fd_set     readFD;
	struct timeval tval = {0,0};
  
#ifdef EXT_MODE
	/* Double check that user parameters are within the range  */
	if ((PORT < 255) || (PORT > 65535)) {
		PRINT_E(stderr, "!!! Error: port number must be within the range: 256 to 65535\n");
		exit(1);
	}
#endif
	
	if (FT < -1) {
		PRINT_E(stderr,"\n!!! Error: finaltime must be a positive, real value or inf\n");
		exit(1);
	}
	
#ifdef VERBOSE	
	else if (FT == RUN_FOREVER) {
        PRINT_M ("\n*** May run forever. Model stop time set to infinity. ***\n");
	}
#endif	

#define P(x) #x
#define OPEN_FIFO(x, FLAG) open("/dev/rtf"P(x), FLAG)

	if ((fd0 = OPEN_FIFO(RTF0, O_RDONLY)) < 0) {
		PRINT_E(stderr, "Error opening /dev/rtf%d\n", RTF0);
		exit(-1);
	}
	if ((fd1 = OPEN_FIFO(RTF1, O_WRONLY)) < 0) {
		PRINT_E(stderr, "Error opening /dev/rtf%d\n", RTF1);
		exit(-1);
	}

	/* Register function 'exit_func' to execute at exit */
	retval = atexit(exit_func);

	/* Check whether there was an error in the init_module() / krnl_main.c */
	if(read(fd0, &cmdHdr, sizeof(cmdHdr)) < 0) {
		PRINT_E(stderr, "\nCannot read from FIFO rtf%d\n", RTF1);
		exit(-1);
	}
	if (cmdHdr.type == ERROR) {
		PRINT_E(stderr, "\nError when initializing the model");
		exit(-1);
	}	

#ifdef EXT_MODE
	
	/* Allocate and initialize structure SendData in shared memory */
	retval = SendDataAlloc();   
	if (retval == ERROR) {
		PRINT_E(stderr, "\nShared memory could not be allocated in user space");
		exit(-1);
	}
	
	{
		const char_T *extParseErrorMsg = ExtParseArgsAndInitUD();
		if (extParseErrorMsg != NULL) {
			PRINT_E(stderr,"\nError processing External Mode command line arguments:\n");
			PRINT_E(stderr,"\t%s", extParseErrorMsg);
			exit(ERROR);
		}
	}
	
	/* Initialise sockets*/
	retval = rt_ExtModeInit();
	if (retval == ERROR) {
		PRINT_E(stderr, "\nSockets could not be initalized.");
		exit(1);
	}
#endif 			

	{
		struct sigaction action;
		
		/* When the Simulink side disconnects,the target machine could still attempt to send data.   *
		 * This causes a system signal SIGPIPE (data being written in a pipe with no recipient)      *
		 * resulting in a termination of the application, when it actually should continue running.  *
		 * To prevent this from happening the signal SIGPIPE is ignored                              */ 
		
		sigemptyset(&action.sa_mask); 
		action.sa_flags     = 0; 
		action.sa_handler   = SIG_IGN;
		action.sa_restorer = NULL;
		retval += sigaction(SIGPIPE, &action, NULL);

		/* Four system signals can make the application stop: 
		   SIGINT:  Control-C (Interrupt character) 
		   SIGQUIT: Control-\ (quit caracter)
		   SIGTSTP: Control-Z (stop signal generated from keyboard)
		   SIGHUP:  terminal closed
		 * The application should not stop when either of the above event happens. Insted when any *
		 * of the key combinations described above occurs the stop_model()  is executed and a soft  *
		 * 'termination'takes place                                                                */

		/* We are leaving the application so block all signals. */
		sigfillset(&action.sa_mask); 
		action.sa_flags     = 0; 
		action.sa_handler   = stop_model;
		action.sa_restorer = NULL;

		retval += sigaction(SIGINT, &action, NULL);
		retval += sigaction(SIGQUIT, &action, NULL);
		retval += sigaction(SIGTSTP, &action, NULL);
		retval += sigaction(SIGHUP, &action, NULL);

		if (retval != 0) {
			PRINT_E(stderr,"\nError occurred when installing a signal handler.");
			exit(1);
		}
	}
  
#if (WAIT_START_MODEL == 0 || !defined(EXT_MODE))
	/* us_main sends the command START_MODEL to krnl_main to make the RT thread periodic  */
	cmdHdr.type = START_MODEL;
	cmdHdr.size = 0;
	if(write(fd1, &cmdHdr, sizeof(cmdHdr)) < 0) {
		PRINT_E(stderr, "\nThe command START_MODEL could not be passed\n ");
		exit(1);
	}
#endif

#  ifdef EXT_MODE
#    define P(x) #x
#    define PRINTF(x) PRINT_M("\n ### Model "P(x)" running in external mode\n")
	  PRINTF(MODEL);
#  else
#    define P(x) #x
#    define PRINTF(x) PRINT_M("\n ### Model "P(x)" running in normal mode\n")
	  PRINTF(MODEL);
#  endif
	
	/* Enter an endless loop that checks whether commands have been sent  *
	 * either from the Simulink  computer (network) or from krnl_main (FIFO) */
	while (1)    {
#ifdef EXT_MODE
		if (rt_UploadServerWork() == ERROR) {
			PRINT_E(stderr, "\n!!! An error occurred in rt_UploadServerWork() !!!\n");
			stop_model();
		}
		if (rt_MsgServerWork() == ERROR) {
			PRINT_E(stderr, "\n!!! An error occurred in rt_MsgServerWork() !!!\n");
			stop_model(); 
		}
#endif

		/* Check whether there is data in the FIFO RTF0 sent from krnl_main. */
		FD_ZERO(&readFD); 
		FD_SET(fd0, &readFD);
		select(fd0 + 1, &readFD, NULL, NULL, &tval); /* check whether there is data in the FIFO RTF0 */
		if (FD_ISSET(fd0, &readFD)){
			if(read(fd0, &cmdHdr, sizeof(cmdHdr))<0)  { 
				PRINT_E(stderr, "\n Command could not be read from FIFO rtf%d\n", RTF1);
				exit(1);
			}
			switch(cmdHdr.type)  {
				case STOP_MODEL:
					exit(0); /* force the function exit_func() to run */
					break;
#ifdef EXT_MODE
				case PROCESS_CONNECT_MSG: 
					retval = ProcessConnectMsg(cmdHdr.size);
					if (retval == ERROR) {
						PRINT_E(stderr, "\n!!! Error in function ProcessConnectMsg() !!!\n");
						stop_model();
					}
					break;
  
				case SEND_DATA:
#  if (WAIT_START_MODEL == 1)
					/* (startModel != 7), either connection, start or arm did'n occurred. Wait for it.  *
					* (startModel & 1), model disconnected while in the loop                           *
					* !(startModel & 8), RT thread has started. No need to wait for start message     */
					while ((startModel != 7) && (startModel & 1) && !(startModel & 8))  {
						rt_ExtModeSleep(0L,100000L); /* Wait for one thenth of a second */
						rt_MsgServerWork();
						rt_UploadServerWork();
					}
					if (startModel == 7) {
						/* Prevent the RT thread from being created again by setin to one bit 4          */
						/* us_main sends the command START_MODEL to krnl_main to make the RT thread periodic  */
						startModel |= 8; 
						cmdHdr.type = START_MODEL;
						cmdHdr.size = 0;
						if(write(fd1, &cmdHdr, sizeof(cmdHdr)) < 0) {
							PRINT_E(stderr,"\nThe command START_MODEL could not be passed\n ");
 							exit(1);
						}
					}
#  endif /* WAIT_START_MODEL */

					/* Free the previous allocation. */
					SendDataBuffFree();
					/* Allocate buffer (SendData->USbuff) in shared memory */
					if (SendDataBuffAlloc() == ERROR) {
						PRINT_E(stderr, "\nShared memory could not be allocated in user space");
						return ERROR;
					}
					break;
#endif /* EXT_MODE */
					
				case DO_DATA_LOG: 
				{
					if (doDataLogging(cmdHdr.size) != SUCCESS)
						PRINT_E(stderr,"\n!!! Creation of MAT-file was was unsuccessful.\n");
					else
						PRINT_M("\nSuccessful creation of MAT-file. ");
				}
					break;
					
				default:
					PRINT_W("\nCommand in FIFO was not recognized\n\tCommand sent: %d\n", cmdHdr.type);
					break;
			}  /* Switch  */
		} 
	}
	return 0;
}
 
