/* Function: sfun_pose2vr ============================================
 * Abstract:
 *   Transforms a DCM to VR rotation vector and angle as needed by VRML
 *
 * Copyright 1998-2001 The MathWorks, Inc.
 * $Revision: 1.2 $
 */
#define S_FUNCTION_NAME  sfun_pose2vr
#define S_FUNCTION_LEVEL 2

#include "simstruc.h"
#include <math.h>

#define DEBUG  0  /* Debugging info 1:on 0:off */

/*************************************************************************
 * Function: transformToVRRotation                                       *
 * Abstract:                                                             *
 *          Takes a transformation matrix and produces the corresponding *
 *          rotation parameters for a VRML scene. It produces a four     *
 *          element vector defining a unit vector and a rotation about   *
 *          that vector in radians.                                      *
 *                                                                       *
 *          rotation = {a1, a2, a3, phi}                                 *
 *                                                                       *
 *          Paul Barnard                                                 *
 *          Copyright (c) 1984-01 by The MathWorks, Inc.                 *
 *************************************************************************/
void transformToVRRotation(SimStruct *S, InputRealPtrsType uT, real_T *rotation)
{
   real_T sigma = 0.0;
   real_T A[3], phi, den;
   real_T T11, T21, T31, T12, T22, T32, T13, T23, T33;

  /* 
   * Make temporary variables T to keep track of indicies.
   * 
   * This algorithm assumes that the transformation matrix
   * coming in is the direction cosine matrix of the body 
   * coordinate system. This tells us the location of the world
   * coordinate system with respect to the body location.
   * Therefore, we need to transpose the incoming 
   * matrix here so we will end up with angle and rotation 
   * of the body with respect to the world.
   *
   * The input matrix is stored column-wise, so we read it
   * in row-wise to get its tranpose.
   */
   T11 = *uT[0];
   T12 = *uT[1];
   T13 = *uT[2];
   T21 = *uT[3];
   T22 = *uT[4];
   T23 = *uT[5];
   T31 = *uT[6];
   T32 = *uT[7];
   T33 = *uT[8];

   /* Calculate an intermediate variable, sigma */
   /* sigma = sum(diag(C)); */
   sigma = T11 + T22 + T33;

   /* If sigma is 3 */
   if (sigma == 3.0) {
      /*
       *  warning('The rotation is undefined')
       *  phi = 0;
       *  A = [1 0 0]';
       */
#if DEBUG
       ssWarning(S, "The rotation is undefined");
#endif
       A[0] = 1.0;
       A[1] = 0.0;
       A[2] = 0.0;
       phi  = 0.0;

   /* ElseIf sigma is -1, */
   } else if (sigma == -1.0) {
      /* Make sure that sums are >= 0. */
       real_T sum0 = (1 + T11)/2.0;
       real_T sum1 = (1 + T22)/2.0;
       real_T sum2 = (1 + T33)/2.0;
       sum0 = sum0 >= 0.0 ? sum0 : 0.0;
       sum1 = sum1 >= 0.0 ? sum1 : 0.0;
       sum2 = sum2 >= 0.0 ? sum2 : 0.0;

      /*
       *  phi = acos((sigma - 1)/2);
       *  A(1) = sqrt((1 + C(1,1))/2);
       *  A(2) = sqrt((1 + C(2,2))/2);
       *  A(3) = sqrt((1 + C(3,3))/2);
       */
       phi = acos((sigma - 1.0)/2.0);
       A[0] = sqrt(sum0);
       A[1] = sqrt(sum1);
       A[2] = sqrt(sum2);

       /* 
        * Resolve sign ambiguity in results here based on these equations.
        * This doesn't need to be done for visulization since we are not
	* concerned if the scene has continuous data from one step to the
	* next.
        *
        * A[0]*A[1] = T12/2;
        * A[1]*A[2] = T23/2;
        * A[2]*A[0] = T31/2;
        */

   /* Else handle the general case */
   } else {
       phi = acos((sigma - 1)/2);
       den = 2*sin(phi);
       A[0] = (T23 - T32)/den;
       A[1] = (T31 - T13)/den;
       A[2] = (T12 - T21)/den;

   }

   /* Return the answer into rotation */
   rotation[0] = A[0];
   rotation[1] = A[1];
   rotation[2] = A[2];
   rotation[3] = phi;

}

/* Function: mdlInitializeSizes ===============================================
 * Abstract:
 *   Setup sizes of the various vectors.
 */
static void mdlInitializeSizes(SimStruct *S)
{
    ssSetNumSFcnParams(S, 0);
    if (ssGetNumSFcnParams(S) != ssGetSFcnParamsCount(S)) {
        return; /* Parameter mismatch will be reported by Simulink */
    }

    /* This block has one inport, expecting a 3x3 matrix given to
     * this s-function as an unoriented 9 element vector; stored
     * columnwise.
     */
    if (!ssSetNumInputPorts(S, 1)) return;
    ssSetInputPortWidth(S, 0, 9);

    /*
     * Set the number of zero-crossing signals to variable here.
     * Later in mdlSetWorkWidths we will explicitly set the number
     * of zero-crossing signals. This must be done because Simulink
     * doesn't have explicit support for setting the number of
     * zero-crossing signals for multi-port s-functions.
     */
    ssSetNumNonsampledZCs(S, 0);

    /* The input port has direct feedthrough */
    ssSetInputPortDirectFeedThrough(S, 0, 1);

    /* This block has one outport of width 4 */
    if (!ssSetNumOutputPorts(S,1)) return;
    ssSetOutputPortWidth(S, 0, 4);

    ssSetNumSampleTimes(S, 1);

    /* Take care when specifying exception free code - see sfuntmpl_doc.c */
    ssSetOptions(S, SS_OPTION_EXCEPTION_FREE_CODE);

}


/* Function: mdlInitializeSampleTimes =========================================
 * Abstract:
 *    Specifiy that we inherit our sample time from the driving block.
 */
static void mdlInitializeSampleTimes(SimStruct *S)
{
    ssSetSampleTime(S, 0, INHERITED_SAMPLE_TIME);
    ssSetOffsetTime(S, 0, FIXED_IN_MINOR_STEP_OFFSET);
}


#undef MDL_START
#if defined(MDL_START)
static void mdlStart(SimStruct *S)
{
}
#endif /* MDL_START */


#undef MDL_ZERO_CROSSINGS
#if defined(MDL_ZERO_CROSSINGS)
#if defined(MATLAB_MEX_FILE) || defined(NRT)
/* Function: mdlZeroCrossings =================================================
 * Abstract:
 *
 */
static void mdlZeroCrossings(SimStruct *S)
{
}
#endif
#endif


/* Function: mdlOutputs =======================================================
 * Abstract:
 *
 *
 */
static void mdlOutputs(SimStruct *S, int_T tid)
{
    InputRealPtrsType uPtr      = ssGetInputPortRealSignalPtrs(S,0);
    real_T            *rotation = ssGetOutputPortRealSignal(S,0);

    transformToVRRotation(S, uPtr, rotation);

}

/* Function: mdlTerminate =====================================================
 * Abstract:
 *    No termination needed, but we are required to have this routine.
 */
static void mdlTerminate(SimStruct *S)
{
}


#undef MDL_RTW
#ifdef MDL_RTW
/* Function: mdlRTW ===========================================================
 * Abstract:
 *	Since we've declared all are parameters as non-tunable, we need
 *	only provide this routine so that they aren't written to the model.rtw
 *	file. The values of the parameters are implicitly encoded in the
 *	sample times.
 */
static void mdlRTW(SimStruct *S)
{
}
#endif /* MDL_RTW */


#ifdef  MATLAB_MEX_FILE    /* Is this file being compiled as a MEX-file? */
#include "simulink.c"      /* MEX-file interface mechanism */
#else
#include "cg_sfun.h"       /* Code generation registration function */
#endif








